import os
from pathlib import Path
from rattler import PrefixRecord, PrefixPaths, PrefixPathsEntry, PrefixPathType, FileMode


def test_load_prefix_record() -> None:
    r = PrefixRecord.from_path(
        Path(__file__).parent / ".." / ".." / ".." / "test-data" / "conda-meta" / "tk-8.6.12-h8ffe710_0.json"
    )
    assert r.arch == "x86_64"
    assert r.build == "h8ffe710_0"
    assert r.build_number == 0
    assert r.channel == "https://conda.anaconda.org/conda-forge/win-64"
    assert len(r.constrains) == 0
    assert len(r.depends) == 2
    assert str(r.extracted_package_dir) == "C:\\Users\\bas\\micromamba\\envs\\conda\\pkgs\\tk-8.6.12-h8ffe710_0"
    assert r.features is None
    assert r.file_name == "tk-8.6.12-h8ffe710_0.tar.bz2"
    assert len(r.files) == len(r.paths_data.paths) == 1099
    paths = r.paths_data
    assert isinstance(paths, PrefixPaths)
    paths_with_placeholder = 0
    assert paths.paths_version == 1
    for entry in paths.paths:
        assert isinstance(entry, PrefixPathsEntry)
        assert entry.relative_path is not None
        assert isinstance(entry.path_type, PrefixPathType)
        if entry.prefix_placeholder is not None:
            paths_with_placeholder += 1
            assert isinstance(entry.file_mode, FileMode)
            assert entry.file_mode.text or entry.file_mode.binary
            assert entry.sha256_in_prefix is not None
        assert entry.sha256 is not None
        assert entry.size_in_bytes > 0

        # check that it implements os.PathLike
        isinstance(entry, os.PathLike)



from __future__ import annotations

from typing import Union

from rattler.version import Version


class VersionWithSource:
    """
    Holds a version and the string it was created from. This is useful if
    you want to retain the original string the version was created from.
    This might be useful in cases where you have multiple strings that
    are represented by the same [`Version`] but you still want to be able to
    distinguish them.

    The string `1.1` and `1.01` represent the same version. When you print
    the parsed version though it will come out as `1.1`. You loose the
    original representation. This class stores the original source string,
    which can be accessed by `source` property.
    """

    def __init__(self, version: Union[str, Version]):
        if not isinstance(version, (str, Version)):
            raise TypeError(
                "VersionWithSource constructor received unsupported type "
                f" {type(version).__name__!r} for the `version` parameter"
            )

        if isinstance(version, str):
            self._source = version
            self._version = Version(version)

        if isinstance(version, Version):
            self._source = str(version)
            self._version = version

    @property
    def version(self) -> Version:
        """
        Returns the `Version` from current object.

        Examples
        --------
        ```python
        >>> v = VersionWithSource("1.01")
        >>> v.version
        Version("1.1")
        >>> v2 = VersionWithSource(v.version)
        >>> v2.version
        Version("1.1")
        >>>
        ```
        """
        return self._version

    @property
    def source(self) -> str:
        """
        Returns the `source` current object was used to create.

        Examples
        --------
        ```python
        >>> v = VersionWithSource("1.01.01")
        >>> v.source
        '1.01.01'
        >>>
        ```
        """
        return self._source

    def __str__(self) -> str:
        """
        Returns the string representation of the version

        Examples
        --------
        ```python
        >>> str(VersionWithSource("1.02.3"))
        '1.02.3'
        >>>
        ```
        """
        return self._source

    def __repr__(self) -> str:
        """
        Returns a representation of the version

        Examples
        --------
        ```python
        >>> VersionWithSource("1.02.3")
        VersionWithSource("1.02.3")
        >>>
        ```
        """
        return f'{type(self).__name__}("{self.__str__()}")'



from rattler.lock.lock_file import LockFile
from rattler.lock.environment import Environment
from rattler.lock.channel import LockChannel
from rattler.lock.hash import PackageHashes
from rattler.lock.package import LockedPackage
from rattler.lock.pypi import PypiPackageData, PypiPackageEnvironmentData

__all__ = [
    "LockFile",
    "Environment",
    "LockChannel",
    "PackageHashes",
    "LockedPackage",
    "PypiPackageData",
    "PypiPackageEnvironmentData",
]



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from rattler.channel.channel import Channel

from rattler.rattler import PyNamelessMatchSpec

if TYPE_CHECKING:
    from rattler.match_spec import MatchSpec
    from rattler.repo_data import PackageRecord


class NamelessMatchSpec:
    """
    Similar to a `MatchSpec` but does not include the package name.
    This is useful in places where the package name is already known
    (e.g. `foo = "3.4.1 *cuda"`).
    """

    def __init__(self, spec: str, strict: bool = False) -> None:
        if isinstance(spec, str):
            self._nameless_match_spec = PyNamelessMatchSpec(spec, strict)
        else:
            raise TypeError(
                "NamelessMatchSpec constructor received unsupported type"
                f" {type(spec).__name__!r} for the 'spec' parameter"
            )

    @property
    def version(self) -> Optional[str]:
        """
        The version spec of the package (e.g. `1.2.3`, `>=1.2.3`, `1.2.*`)
        """
        return self._nameless_match_spec.version

    @property
    def build(self) -> Optional[str]:
        """
        The build string of the package (e.g. `py37_0`, `py37h6de7cb9_0`, `py*`)
        """
        return self._nameless_match_spec.build

    @property
    def build_number(self) -> Optional[str]:
        """
        The build number of the package.
        """
        return self._nameless_match_spec.build_number

    @property
    def file_name(self) -> Optional[str]:
        """
        Match the specific filename of the package.
        """
        return self._nameless_match_spec.file_name

    @property
    def channel(self) -> Optional[Channel]:
        """
        The channel of the package.
        """
        if (channel := self._nameless_match_spec.channel) is not None:
            return Channel(channel.name)
        return None

    @property
    def subdir(self) -> Optional[str]:
        """
        The subdir of the channel.
        """
        return self._nameless_match_spec.subdir

    @property
    def namespace(self) -> Optional[str]:
        """
        The namespace of the package.
        """
        return self._nameless_match_spec.namespace

    @property
    def md5(self) -> Optional[bytes]:
        """
        The md5 hash of the package.
        """
        return self._nameless_match_spec.md5

    @property
    def sha256(self) -> Optional[bytes]:
        """
        The sha256 hash of the package.
        """
        return self._nameless_match_spec.sha256

    def matches(self, package_record: PackageRecord) -> bool:
        """
        Match a MatchSpec against a PackageRecord
        """
        return self._nameless_match_spec.matches(package_record._record)

    @classmethod
    def _from_py_nameless_match_spec(cls, py_nameless_match_spec: PyNamelessMatchSpec) -> NamelessMatchSpec:
        """
        Construct py-rattler NamelessMatchSpec from PyNamelessMatchSpec FFI object.
        """
        nameless_match_spec = cls.__new__(cls)
        nameless_match_spec._nameless_match_spec = py_nameless_match_spec

        return nameless_match_spec

    @classmethod
    def from_match_spec(cls, spec: MatchSpec) -> NamelessMatchSpec:
        """
        Constructs a NamelessMatchSpec from a MatchSpec.

        Examples
        --------
        ```python
        >>> from rattler import MatchSpec
        >>> NamelessMatchSpec.from_match_spec(MatchSpec("foo ==3.4"))
        NamelessMatchSpec("==3.4")
        >>>
        ```
        """
        return cls._from_py_nameless_match_spec(PyNamelessMatchSpec.from_match_spec(spec._match_spec))

    def __str__(self) -> str:
        """
        Returns a string representation of the NamelessMatchSpec.

        Examples
        --------
        ```python
        >>> str(NamelessMatchSpec("3.4"))
        '==3.4'
        >>>
        ```
        """
        return self._nameless_match_spec.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the NamelessMatchSpec.

        Examples
        --------
        ```python
        >>> NamelessMatchSpec("3.4")
        NamelessMatchSpec("==3.4")
        >>>
        ```
        """
        return f'NamelessMatchSpec("{self._nameless_match_spec.as_str()}")'



from __future__ import annotations
from typing import Optional
from rattler.rattler import PyPackageHashes


class PackageHashes:
    _hashes: PyPackageHashes

    @property
    def md5(self) -> Optional[bytes]:
        """
        Returns the Sha256 hash.
        """
        return self._hashes.md5

    @property
    def sha256(self) -> Optional[bytes]:
        """
        Returns the Sha256 hash.
        """
        return self._hashes.sha256

    @classmethod
    def _from_py_package_hashes(cls, pkg_hashes: PyPackageHashes) -> PackageHashes:
        """
        Construct Rattler PackageHashes from FFI PyPackageHashes object.
        """
        hashes = cls.__new__(cls)
        hashes._hashes = pkg_hashes
        return hashes

    def __repr__(self) -> str:
        return "PackageHashes()"



from __future__ import annotations
import os
from typing import List, Optional
from rattler.match_spec.match_spec import MatchSpec
from rattler.package.package_name import PackageName

from rattler.rattler import PyRecord
from rattler.version.version import Version


class PackageRecord:
    """
    A single record in the Conda repodata. A single
    record refers to a single binary distribution
    of a package on a Conda channel.
    """

    _record: PyRecord

    def matches(self, spec: MatchSpec) -> bool:
        """
        Match a [`PackageRecord`] against a [`MatchSpec`].
        """
        return spec.matches(self)

    @staticmethod
    def from_index_json(
        path: os.PathLike[str],
        size: Optional[int] = None,
        sha256: Optional[str] = None,
        md5: Optional[str] = None,
    ) -> PackageRecord:
        """
        Builds a PackageRecord from an `index.json`.
        These can be found in `info` directory inside an
        extracted package archive.

        Examples
        --------
        ```python
        >>> record = PackageRecord.from_index_json(
        ...     "../test-data/conda-meta/pysocks-1.7.1-pyh0701188_6.json"
        ... )
        >>> assert isinstance(record, PackageRecord)
        >>>
        ```
        """
        return PackageRecord._from_py_record(PyRecord.from_index_json(path, size, sha256, md5))

    @staticmethod
    def sort_topologically(records: List[PackageRecord]) -> List[PackageRecord]:
        """
        Sorts the records topologically.
        This function is deterministic, meaning that it will return the same result
        regardless of the order of records and of the depends vector inside the records.
        Note that this function only works for packages with unique names.

        Examples
        --------
        ```python
        >>> from os import listdir
        >>> from os.path import isfile, join
        >>> from rattler import PrefixRecord
        >>> records = [
        ...     PrefixRecord.from_path(join("../test-data/conda-meta/", f))
        ...     for f in listdir("../test-data/conda-meta")
        ...     if isfile(join("../test-data/conda-meta", f))
        ... ]
        >>> sorted = PackageRecord.sort_topologically(records)
        >>> sorted[0].name
        PackageName("vc")
        >>>
        ```
        """
        return [PackageRecord._from_py_record(p) for p in PyRecord.sort_topologically(records)]

    @classmethod
    def _from_py_record(cls, py_record: PyRecord) -> PackageRecord:
        """
        Construct Rattler PackageRecord from FFI PyRecord object.
        """

        # quick sanity check
        assert py_record.is_package_record
        record = cls.__new__(cls)
        record._record = py_record
        return record

    @property
    def arch(self) -> Optional[str]:
        """
        Optionally the architecture the package supports.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.arch
        'x86_64'
        >>>
        ```
        """
        return self._record.arch

    @property
    def build(self) -> str:
        """
        The build string of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.build
        'hcfcfb64_0'
        >>>
        ```
        """
        return self._record.build

    @property
    def build_number(self) -> int:
        """
        The build number of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.build_number
        0
        >>>
        ```
        """
        return self._record.build_number

    @property
    def constrains(self) -> List[str]:
        """
        Additional constraints on packages.
        Constrains are different from depends in that packages
        specified in depends must be installed next to this package,
        whereas packages specified in constrains are not required to
        be installed, but if they are installed they must follow
        these constraints.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.constrains
        []
        >>>
        ```
        """
        return self._record.constrains

    @property
    def depends(self) -> List[str]:
        """
        Specification of packages this package depends on.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.depends
        ['ucrt >=10.0.20348.0', 'vc >=14.2,<15', 'vs2015_runtime >=14.29.30139']
        >>>
        ```
        """
        return self._record.depends

    @property
    def features(self) -> Optional[str]:
        """
        Features are a deprecated way to specify different feature
        sets for the conda solver. This is not supported anymore and
        should not be used. Instead, mutex packages should be used
        to specify mutually exclusive features.
        """
        return self._record.features

    @property
    def legacy_bz2_md5(self) -> Optional[str]:
        """
        A deprecated md5 hash.
        """
        return self._record.legacy_bz2_md5

    @property
    def legacy_bz2_size(self) -> Optional[int]:
        """
        A deprecated package archive size.
        """
        return self._record.legacy_bz2_size

    @property
    def license(self) -> Optional[str]:
        """
        The specific license of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.license
        'Unlicense'
        >>>
        ```
        """
        return self._record.license

    @property
    def license_family(self) -> Optional[str]:
        """
        The license family.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/pip-23.0-pyhd8ed1ab_0.json"
        ... )
        >>> record.license_family
        'MIT'
        >>>
        ```

        """
        return self._record.license_family

    @property
    def md5(self) -> Optional[str]:
        """
        Optionally a MD5 hash of the package archive.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.md5
        '5E5A97795DE72F8CC3BAF3D9EA6327A2'
        >>>
        ```
        """
        return self._record.md5

    @property
    def name(self) -> PackageName:
        """
        The name of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.name
        PackageName("libsqlite")
        >>>
        ```
        """
        return PackageName._from_py_package_name(self._record.name)

    @property
    def platform(self) -> Optional[str]:
        """
        Optionally the platform the package supports.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.platform
        'win32'
        >>>
        ```
        """
        return self._record.platform

    @property
    def sha256(self) -> Optional[str]:
        """
        Optionally a SHA256 hash of the package archive.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.sha256
        '4E50B3D90A351C9D47D239D3F90FCE4870DF2526E4F7FEF35203AB3276A6DFC9'
        >>>
        """
        return self._record.sha256

    @property
    def size(self) -> Optional[int]:
        """
        Optionally the size of the package archive in bytes.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.size
        669941
        >>>
        ```
        """
        return self._record.size

    @property
    def subdir(self) -> str:
        """
        The subdirectory where the package can be found.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.subdir
        'win-64'
        >>>
        ```
        """
        return self._record.subdir

    @property
    def timestamp(self) -> Optional[int]:
        """
        The date this entry was created.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.timestamp
        1668697639
        >>>
        ```
        """
        return self._record.timestamp

    @property
    def track_features(self) -> List[str]:
        """
        Track features are nowadays only used to downweight
        packages (ie. give them less priority).
        To that effect, the number of track features is
        counted (number of commas) and the package is downweighted
        by the number of track_features.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.track_features
        []
        >>>
        ```
        """
        return self._record.track_features

    @property
    def version(self) -> Version:
        """
        The version of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.version
        Version("3.40.0")
        >>>
        ```
        """
        return Version._from_py_version(self._record.version)

    def __str__(self) -> str:
        """
        Returns the string representation of the PackageRecord.

        Examples
        --------
        ```python
        >>> record = PackageRecord.from_index_json(
        ...     "../test-data/conda-meta/pysocks-1.7.1-pyh0701188_6.json"
        ... )
        >>> str(record)
        'pysocks=1.7.1=pyh0701188_6'
        >>>
        ```
        """
        return self._record.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the PackageRecord.

        Examples
        --------
        ```python
        >>> record = PackageRecord.from_index_json(
        ...     "../test-data/conda-meta/pysocks-1.7.1-pyh0701188_6.json"
        ... )
        >>> record
        PackageRecord("pysocks=1.7.1=pyh0701188_6")
        >>>
        ```
        """
        return f'PackageRecord("{self.__str__()}")'



from rattler.version import Version, VersionWithSource
from rattler.match_spec import MatchSpec, NamelessMatchSpec
from rattler.repo_data import (
    PackageRecord,
    RepoData,
    RepoDataRecord,
    PatchInstructions,
    SparseRepoData,
)
from rattler.channel import Channel, ChannelConfig, ChannelPriority
from rattler.networking import AuthenticatedClient, fetch_repo_data
from rattler.virtual_package import GenericVirtualPackage, VirtualPackage
from rattler.package import (
    PackageName,
    AboutJson,
    RunExportsJson,
    PathsJson,
    PathsEntry,
    PathType,
    PrefixPlaceholder,
    FileMode,
)
from rattler.prefix import PrefixRecord, PrefixPaths, PrefixPathsEntry, PrefixPathType
from rattler.solver import solve
from rattler.platform import Platform
from rattler.utils.rattler_version import get_rattler_version as _get_rattler_version
from rattler.linker import link
from rattler.index import index
from rattler.lock import (
    LockFile,
    Environment,
    LockChannel,
    PackageHashes,
    LockedPackage,
    PypiPackageData,
    PypiPackageEnvironmentData,
)

__version__ = _get_rattler_version()
del _get_rattler_version

__all__ = [
    "Version",
    "VersionWithSource",
    "MatchSpec",
    "NamelessMatchSpec",
    "PackageRecord",
    "Channel",
    "ChannelConfig",
    "ChannelPriority",
    "AuthenticatedClient",
    "PatchInstructions",
    "RepoDataRecord",
    "RepoData",
    "fetch_repo_data",
    "GenericVirtualPackage",
    "VirtualPackage",
    "PackageName",
    "PrefixRecord",
    "PrefixPaths",
    "PrefixPathsEntry",
    "PrefixPathType",
    "SparseRepoData",
    "LockFile",
    "Environment",
    "LockChannel",
    "PackageHashes",
    "LockedPackage",
    "PypiPackageData",
    "PypiPackageEnvironmentData",
    "solve",
    "Platform",
    "link",
    "index",
    "AboutJson",
    "RunExportsJson",
    "PathsJson",
    "PathsEntry",
    "PathType",
    "PrefixPlaceholder",
    "FileMode",
    "IndexJson",
]



from __future__ import annotations

from rattler.rattler import PyPackageName


class PackageName:
    def __init__(self, source: str) -> None:
        if not isinstance(source, str):
            raise TypeError(
                "PackageName constructor received unsupported type "
                f" {type(source).__name__!r} for the `source` parameter"
            )
        self._name = PyPackageName(source)

    @staticmethod
    def unchecked(normalized: str) -> PackageName:
        """
        Constructs a new `PackageName` from a string without checking if the string is actually a
        valid or normalized conda package name. This should only be used if you are sure that the
        input string is valid.

        Examples
        --------
        ```python
        >>> p = PackageName.unchecked("test_xyz")
        >>>
        ```
        """
        return PackageName._from_py_package_name(PyPackageName.new_unchecked(normalized))

    @classmethod
    def _from_py_package_name(cls, py_package_name: PyPackageName) -> PackageName:
        """Construct Rattler PackageName from FFI PyPackageName object."""
        package_name = cls.__new__(cls)
        package_name._name = py_package_name
        return package_name

    @property
    def source(self) -> str:
        """
        Returns the source representation of the package name.
        This is the string from which this instance was created.

        Examples
        --------
        ```python
        >>> p = PackageName("test-xyz")
        >>> p.source
        'test-xyz'
        >>>
        ```
        """
        return self._name.source

    @property
    def normalized(self) -> str:
        """
        Returns the normalized version of the package name.
        The normalized string is guaranteed to be a valid conda package name.

        Examples
        --------
        ```python
        >>> p = PackageName("test-xyz")
        >>> p.normalized
        'test-xyz'
        >>>
        ```
        """
        return self._name.normalized

    def __hash__(self) -> int:
        """
        Computes the hash of this instance.

        Examples
        --------
        ```python
        >>> hash(PackageName("test-abc")) == hash(PackageName("test-abc"))
        True
        >>> hash(PackageName("test-abc")) == hash(PackageName("test-ABC"))
        True
        >>> hash(PackageName("test-abc")) == hash(PackageName("abc-test"))
        False
        >>>
        ```
        """
        return self._name.__hash__()

    def __eq__(self, other: object) -> bool:
        """
        Returns True if this instance represents the same PackageName as `other`.

        Examples
        --------
        ```python
        >>> PackageName("test-abc") == PackageName("abc-test")
        False
        >>> PackageName("test-abc") == PackageName("test-abc")
        True
        >>> PackageName("test-abc") == PackageName("test-ABC")
        True
        >>> PackageName("test-abc") == "test-abc"
        False
        >>>
        ```
        """
        if not isinstance(other, PackageName):
            return False

        return self._name == other._name

    def __ne__(self, other: object) -> bool:
        """
        Returns True if this instance does not represents the same PackageName as `other`.

        Examples
        --------
        ```python
        >>> PackageName("test-abc") != PackageName("test-abc")
        False
        >>> PackageName("test-abc") != PackageName("test-ABC")
        False
        >>> PackageName("test-abc") != PackageName("abc-test")
        True
        >>> PackageName("test-abc") != "test-abc"
        True
        >>>
        ```
        """
        if not isinstance(other, PackageName):
            return True

        return self._name != other._name

    def __repr__(self) -> str:
        """
        Returns a representation of the PackageName.

        Examples
        --------
        ```python
        >>> p = PackageName("test-xyz")
        >>> p
        PackageName("test-xyz")
        >>>
        ```
        """
        return f'PackageName("{self.source}")'



# type: ignore
import os.path

from rattler import (
    solve,
    Channel,
    ChannelPriority,
    MatchSpec,
    RepoDataRecord,
    SparseRepoData,
)


def test_solve():
    linux64_chan = Channel("conda-forge")
    data_dir = os.path.join(os.path.dirname(__file__), "../../../test-data/")
    linux64_path = os.path.join(data_dir, "channels/conda-forge/linux-64/repodata.json")
    linux64_data = SparseRepoData(
        channel=linux64_chan,
        subdir="linux-64",
        path=linux64_path,
    )

    solved_data = solve(
        [MatchSpec("python"), MatchSpec("sqlite")],
        [linux64_data],
    )

    assert isinstance(solved_data, list)
    assert isinstance(solved_data[0], RepoDataRecord)
    assert len(solved_data) == 19


def test_solve_channel_priority_disabled():
    cf_chan = Channel("conda-forge")
    data_dir = os.path.join(os.path.dirname(__file__), "../../../test-data/")
    cf_path = os.path.join(data_dir, "channels/conda-forge/linux-64/repodata.json")
    cf_data = SparseRepoData(
        channel=cf_chan,
        subdir="linux-64",
        path=cf_path,
    )

    pytorch_chan = Channel("pytorch")
    pytorch_path = os.path.join(data_dir, "channels/pytorch/linux-64/repodata.json")
    pytorch_data = SparseRepoData(
        channel=pytorch_chan,
        subdir="linux-64",
        path=pytorch_path,
    )

    solved_data = solve(
        [MatchSpec("pytorch-cpu=0.4.1=py36_cpu_1")],
        [cf_data, pytorch_data],
        channel_priority=ChannelPriority.Disabled,
    )

    assert isinstance(solved_data, list)
    assert isinstance(solved_data[0], RepoDataRecord)
    assert list(filter(lambda r: r.file_name.startswith("pytorch-cpu-0.4.1-py36_cpu_1"), solved_data))[0].channel == "https://conda.anaconda.org/pytorch/"
    assert len(solved_data) == 32



from rattler.repo_data.package_record import PackageRecord
from rattler.repo_data.repo_data import RepoData
from rattler.repo_data.patch_instructions import PatchInstructions
from rattler.repo_data.record import RepoDataRecord
from rattler.repo_data.sparse import SparseRepoData

__all__ = [
    "PackageRecord",
    "RepoData",
    "PatchInstructions",
    "RepoDataRecord",
    "SparseRepoData",
]



from rattler.package.package_name import PackageName
from rattler.package.about_json import AboutJson
from rattler.package.run_exports_json import RunExportsJson
from rattler.package.paths_json import (
    PathsJson,
    PathsEntry,
    PathType,
    PrefixPlaceholder,
    FileMode,
)
from rattler.package.index_json import IndexJson

__all__ = [
    "PackageName",
    "AboutJson",
    "RunExportsJson",
    "PathsJson",
    "PathsEntry",
    "PathType",
    "PrefixPlaceholder",
    "FileMode",
    "IndexJson",
]



from __future__ import annotations
import os
from pathlib import Path
from typing import List, Optional
from rattler.rattler import (
    PyPathsJson,
    PyPathsEntry,
    PyPathType,
    PyPrefixPlaceholder,
    PyFileMode,
)


class PathsJson:
    """
    A representation of the `paths.json` file found in package archives.
    The `paths.json` file contains information about every file included with the package.
    """

    _inner: PyPathsJson

    @staticmethod
    def from_package_archive(path: os.PathLike[str]) -> PathsJson:
        """
        Parses the package file from archive.
        Note: If you want to extract multiple `info/*` files then this will be slightly
              slower than manually iterating over the archive entries with
              custom logic as this skips over the rest of the archive

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> paths_json
        PathsJson()
        >>>
        ```
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_package_archive(path))

    @staticmethod
    def from_path(path: os.PathLike[str]) -> PathsJson:
        """
        Parses the object from a file specified by a `path`, using a format
        appropriate for the file type.

        For example, if the file is in JSON format, this function reads the data
        from the file at the specified path, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not read, this function returns an error.
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_path(Path(path)))

    @staticmethod
    def from_package_directory(path: os.PathLike[str]) -> PathsJson:
        """
        Parses the object by looking up the appropriate file from the root of the
        specified Conda archive directory, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function reads the
        appropriate file from the archive, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not be read, this function returns an error.
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_package_directory(Path(path)))

    @staticmethod
    def from_str(string: str) -> PathsJson:
        """
        Parses the object from a string, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function parses the JSON
        string and returns the resulting object. If the file is not in a parsable
        format, this function returns an error.
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_str(string))

    @staticmethod
    def package_path() -> str:
        """
        Returns the path to the file within the Conda archive.

        The path is relative to the root of the archive and includes any necessary
        directories.
        """
        return PathsJson.package_path()

    @staticmethod
    def from_deprecated_package_directory(path: os.PathLike[str]) -> PathsJson:
        """
        Constructs a new instance by reading older (deprecated) files from a package directory.

        In older package archives the `paths.json` file does not exist. These packages contain the
        information normally present in the `paths.json` file spread over different files in the
        archive.

        This function reads the different files and tries to reconstruct a `paths.json` from it.
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_deprecated_package_directory(path))

    @staticmethod
    def from_package_directory_with_deprecated_fallback(
        path: os.PathLike[str],
    ) -> PathsJson:
        """
        Reads the file from a package archive directory. If the `paths.json` file could not be found
        use the `from_deprecated_package_directory` method as a fallback.
        """
        return PathsJson._from_py_paths_json(PyPathsJson.from_package_directory_with_deprecated_fallback(path))

    @property
    def paths(self) -> List[PathsEntry]:
        """
        All entries included in the package.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> paths_json.paths
        [PathsEntry(), ...]
        >>>
        ```
        """
        return [PathsEntry._from_py_paths_entry(path) for path in self._inner.paths]

    @property
    def paths_version(self) -> int:
        """
        The version of the file.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> paths_json.paths_version
        1
        >>>
        ```
        """
        return self._inner.paths_version

    @classmethod
    def _from_py_paths_json(cls, py_paths_json: PyPathsJson) -> PathsJson:
        paths_json = cls.__new__(cls)
        paths_json._inner = py_paths_json

        return paths_json

    def __repr__(self) -> str:
        """
        Returns a representation of the PathsJson.
        """
        return "PathsJson()"


class PathsEntry:
    """
    A single entry in the `paths.json` file.
    """

    _inner: PyPathsEntry

    @property
    def relative_path(self) -> str:
        """
        The relative path from the root of the package.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.relative_path
        'bin/2to3'
        >>>
        ```
        """
        return self._inner.relative_path

    @property
    def no_link(self) -> bool:
        """
        Whether or not this file should be linked or not when installing the package.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.no_link
        False
        >>>
        ```
        """
        return self._inner.no_link

    @property
    def path_type(self) -> PathType:
        """
        Determines how to include the file when installing the package.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.path_type
        PathType()
        >>>
        ```
        """
        return PathType._from_py_path_type(self._inner.path_type)

    @property
    def prefix_placeholder(self) -> Optional[PrefixPlaceholder]:
        """
        Optionally the placeholder prefix used in the file. If this value is `None`
        the prefix is not present in the file.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.prefix_placeholder
        >>>
        ```
        """
        if placeholder := self._inner.prefix_placeholder:
            return PrefixPlaceholder._from_py_prefix_placeholder(placeholder)

        return None

    @property
    def sha256(self) -> Optional[bytes]:
        """
        A hex representation of the SHA256 hash of the contents of the file.
        This entry is only present in version 1 of the paths.json file.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.sha256.hex()
        '863589436be82be20e3efbe0e36f20510747b1892b3cbcb398c4c48ad7e96fcc'
        >>>
        ```
        """
        return self._inner.sha256

    @property
    def size_in_bytes(self) -> Optional[int]:
        """
        The size of the file in bytes.
        This entry is only present in version 1 of the paths.json file.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> entry.size_in_bytes
        347
        >>>
        ```
        """
        if size := self._inner.size_in_bytes:
            return size

        return None

    @classmethod
    def _from_py_paths_entry(cls, py_paths_entry: PyPathsEntry) -> PathsEntry:
        paths_entry = cls.__new__(cls)
        paths_entry._inner = py_paths_entry

        return paths_entry

    def __repr__(self) -> str:
        """
        Returns a representation of the PathsEntry.
        """
        return "PathsEntry()"


class PathType:
    """
    The path type of the path entry
    """

    _inner: PyPathType

    @property
    def hardlink(self) -> bool:
        """
        The path should be hard linked (the default).

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> path_type = entry.path_type
        >>> path_type.hardlink
        False
        >>>
        ```
        """
        return self._inner.hardlink

    @property
    def softlink(self) -> bool:
        """
        The path should be soft linked.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> path_type = entry.path_type
        >>> path_type.softlink
        True
        >>>
        ```
        """
        return self._inner.softlink

    @property
    def directory(self) -> bool:
        """
        This should explicitly create an empty directory.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> entry = paths_json.paths[0]
        >>> path_type = entry.path_type
        >>> path_type.directory
        False
        >>>
        ```
        """
        return self._inner.directory

    @classmethod
    def _from_py_path_type(cls, py_paths_type: PyPathType) -> PathType:
        path_type = cls.__new__(cls)
        path_type._inner = py_paths_type

        return path_type

    def __repr__(self) -> str:
        """
        Returns a representation of the PathType.
        """
        return "PathType()"


class PrefixPlaceholder:
    """
    Description off a placeholder text found in a file that must be replaced
    when installing the file into the prefix.
    """

    _inner: PyPrefixPlaceholder

    @property
    def file_mode(self) -> FileMode:
        """
        The type of the file, either binary or text.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/conda-22.9.0-py38haa244fe_2.tar.bz2"
        ... )
        >>> entry = paths_json.paths[-1]
        >>> entry.prefix_placeholder.file_mode
        FileMode()
        >>>
        ```
        """
        return FileMode._from_py_file_mode(self._inner.file_mode)

    @property
    def placeholder(self) -> str:
        """
        The placeholder prefix used in the file. This is the path of the
        prefix when the package was build.

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/conda-22.9.0-py38haa244fe_2.tar.bz2"
        ... )
        >>> entry = paths_json.paths[-1]
        >>> entry.prefix_placeholder.placeholder
        'D:\\\\bld\\\\conda_1667595064120\\\\_h_env'
        >>>
        ```
        """
        return self._inner.placeholder

    @classmethod
    def _from_py_prefix_placeholder(cls, py_prefix_placeholder: PyPrefixPlaceholder) -> PrefixPlaceholder:
        prefix_placeholder = cls.__new__(cls)
        prefix_placeholder._inner = py_prefix_placeholder

        return prefix_placeholder

    def __repr__(self) -> str:
        """
        Returns a representation of the PrefixPlaceholder.
        """
        return "PrefixPlaceholder()"


class FileMode:
    """
    The file mode of the entry.
    """

    _inner: PyFileMode

    @property
    def binary(self) -> bool:
        """
        The file is a binary file (needs binary prefix replacement).

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/conda-22.9.0-py38haa244fe_2.tar.bz2"
        ... )
        >>> entry = paths_json.paths[-1]
        >>> file_mode = entry.prefix_placeholder.file_mode
        >>> file_mode.binary
        False
        >>>
        ```
        """
        return self._inner.binary

    @property
    def text(self) -> bool:
        """
        The file is a text file (needs text prefix replacement).

        Examples
        --------
        ```python
        >>> paths_json = PathsJson.from_package_archive(
        ...     "../test-data/conda-22.9.0-py38haa244fe_2.tar.bz2"
        ... )
        >>> entry = paths_json.paths[-1]
        >>> file_mode = entry.prefix_placeholder.file_mode
        >>> file_mode.text
        True
        >>>
        ```
        """
        return self._inner.text

    @classmethod
    def _from_py_file_mode(cls, py_file_mode: PyFileMode) -> FileMode:
        file_mode = cls.__new__(cls)
        file_mode._inner = py_file_mode

        return file_mode

    def __repr__(self) -> str:
        """
        Returns a representation of the FileMode.
        """
        return "FileMode()"



from __future__ import annotations
from typing import Callable, List, Optional, Union, TYPE_CHECKING


from rattler.rattler import py_fetch_repo_data
from rattler.repo_data.sparse import SparseRepoData

if TYPE_CHECKING:
    import os
    from rattler.channel import Channel
    from rattler.platform import Platform


async def fetch_repo_data(
    *,
    channels: List[Channel],
    platforms: List[Platform],
    cache_path: Union[str, os.PathLike[str]],
    callback: Optional[Callable[[int, int], None]],
) -> List[SparseRepoData]:
    """
    Returns a list of RepoData for given channels and platform.

    Arguments:
        channels: A list of `Channel`s to fetch repo data.
        platforms: A list of `Platform`s for which the repo data
                   should be fetched.
        cache_path: A `os.PathLike[str]` where the repo data should
                    be downloaded.
        callback: A `Callable[[int, int], None]` to report the download
                  progress of repo data.

    Returns:
        A list of `SparseRepoData` for requested channels and platforms.
    """
    repo_data_list = await py_fetch_repo_data(
        [channel._channel for channel in channels],
        [platform._inner for platform in platforms],
        cache_path,
        callback,
    )

    return [SparseRepoData._from_py_sparse_repo_data(repo_data) for repo_data in repo_data_list]



from __future__ import annotations
from typing import Optional, Tuple
from rattler.lock.pypi import PypiPackageData, PypiPackageEnvironmentData

from rattler.rattler import PyLockedPackage
from rattler.repo_data.record import RepoDataRecord


class LockedPackage:
    _package: PyLockedPackage

    def as_conda(self) -> Optional[RepoDataRecord]:
        """
        Returns this instance as a [`RepoDataRecord`] if this instance represents a conda package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_packages = env.packages(Platform("osx-arm64"))
        >>> conda_pkg = [pkg for pkg in lock_packages if pkg.is_conda]
        >>> conda_pkg[0].as_conda()
        RepoDataRecord(...)
        >>>
        ```
        """
        if pkg := self._package.as_conda():
            return RepoDataRecord._from_py_record(pkg)
        return None

    def as_pypi(self) -> Optional[Tuple[PypiPackageData, PypiPackageEnvironmentData]]:
        """
        Returns this instance as a tuple of [`PypiPackageData`] & [`PypiPackageEnvironmentData`]
        if this instance represents a pypi package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_packages = env.packages(Platform("osx-arm64"))
        >>> pypi_pkg = [pkg for pkg in lock_packages if pkg.is_pypi]
        >>> pypi_pkg[0].as_pypi()
        (PypiPackageData(), PypiPackageEnvironmentData())
        >>>
        ```
        """
        if pkg := self._package.as_pypi():
            (pkg_data, env_data) = pkg
            return (
                PypiPackageData._from_py_pypi_package_data(pkg_data),
                PypiPackageEnvironmentData._from_py_pypi_env_data(env_data),
            )
        return None

    @property
    def is_conda(self) -> bool:
        """
        Returns true if this package represents a conda package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_packages = env.packages(Platform("osx-arm64"))
        >>> conda_pkgs = [pkg for pkg in lock_packages if pkg.url_or_path.startswith("https://conda.anaconda.org/")]
        >>> conda_pkgs[0].is_conda
        True
        >>>
        ```
        """
        return self._package.is_conda

    @property
    def is_pypi(self) -> bool:
        """
        Returns true if this package represents a conda package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_packages = env.packages(Platform("osx-arm64"))
        >>> pypi_pkgs = [pkg for pkg in lock_packages if pkg.url_or_path.startswith("https://files.pythonhosted.org/")]
        >>> pypi_pkgs[0].is_pypi
        True
        >>>
        ```
        """
        return self._package.is_pypi

    @property
    def name(self) -> str:
        """
        Returns the name of the package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_package = env.packages(Platform("osx-arm64"))[0]
        >>> lock_package.name
        'tzdata'
        >>>
        ```
        """
        return self._package.name

    @property
    def url_or_path(self) -> str:
        """
        Returns the URL of the package

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_package = env.packages(Platform("osx-arm64"))[0]
        >>> lock_package.url_or_path
        'https://conda.anaconda.org/...'
        >>>
        ```
        """
        return self._package.url_or_path

    @property
    def version(self) -> str:
        """
        Returns the version string of the package.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> lock_package = env.packages(Platform("osx-arm64"))[0]
        >>> lock_package.version
        '2024a'
        >>>
        ```
        """
        return self._package.version

    @classmethod
    def _from_py_lock_package(cls, lock_pkg: PyLockedPackage) -> LockedPackage:
        """
        Construct Rattler LockedPackage from FFI PyLockedPackage object.
        """
        pkg = cls.__new__(cls)
        pkg._package = lock_pkg
        return pkg

    def __repr__(self) -> str:
        """
        Returns a representation of the LockedPackage.
        """
        return "LockedPackage()"



from __future__ import annotations

from rattler.version import Version
from rattler.package import PackageName

from rattler.rattler import PyGenericVirtualPackage


class GenericVirtualPackage:
    _generic_virtual_package: PyGenericVirtualPackage

    def __init__(self, name: PackageName, version: Version, build_string: str) -> None:
        if not isinstance(name, PackageName):
            raise TypeError(
                "GenericVirtualPackage constructor received unsupported type "
                f" {type(name).__name__!r} for the `name` parameter"
            )

        if not isinstance(version, Version):
            raise TypeError(
                "GenericVirtualPackage constructor received unsupported type "
                f" {type(version).__name__!r} for the `version` parameter"
            )

        if not isinstance(build_string, str):
            raise TypeError(
                "GenericVirtualPackage constructor received unsupported type "
                f" {type(build_string).__name__!r} for the `build_string` parameter"
            )

        self._generic_virtual_package = PyGenericVirtualPackage(name._name, version._version, build_string)

    @property
    def name(self) -> PackageName:
        """
        Returns the name of the package

        Examples
        --------
        ```python
        >>> from rattler.package.package_name import PackageName
        >>> from rattler.version.version import Version
        >>> gvp = GenericVirtualPackage(PackageName("__archspec"), Version("1"), "x86_64")
        >>> gvp.name
        PackageName("__archspec")
        >>> gvp.name.source
        '__archspec'
        >>> gvp.name.normalized
        '__archspec'
        >>>
        ```
        """
        return PackageName._from_py_package_name(self._generic_virtual_package.name)

    @property
    def version(self) -> Version:
        """
        Returns the version of the package

        Examples
        --------
        ```python
        >>> from rattler.package.package_name import PackageName
        >>> from rattler.version.version import Version
        >>> gvp = GenericVirtualPackage(PackageName("__archspec"), Version("1"), "x86_64")
        >>> gvp.version
        Version("1")
        >>>
        ```
        """
        return Version._from_py_version(self._generic_virtual_package.version)

    @property
    def build_string(self) -> str:
        """
        Returns the build identifier of the package.

        Examples
        --------
        ```python
        >>> from rattler.package.package_name import PackageName
        >>> from rattler.version.version import Version
        >>> gvp = GenericVirtualPackage(PackageName("__archspec"), Version("1"), "x86_64")
        >>> gvp.build_string
        'x86_64'
        >>>
        ```
        """
        return self._generic_virtual_package.build_string

    @classmethod
    def _from_py_generic_virtual_package(
        cls, py_generic_virtual_package: PyGenericVirtualPackage
    ) -> GenericVirtualPackage:
        """
        Construct Rattler GenericVirtualPackage from FFI
        PyGenericVirtualPackage object.
        """
        generic_virtual_package = cls.__new__(cls)
        generic_virtual_package._generic_virtual_package = py_generic_virtual_package
        return generic_virtual_package

    def __str__(self) -> str:
        """
        Returns the string representation of the GenericVirtualPackage

        Examples
        --------
        ```python
        >>> from rattler.package.package_name import PackageName
        >>> from rattler.version.version import Version
        >>> gvp = GenericVirtualPackage(PackageName("__archspec"), Version("1"), "x86_64")
        >>> str(gvp)
        '__archspec=1=x86_64'
        >>>
        ```
        """

        return self._generic_virtual_package.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the GenericVirtualPackage

        Examples
        --------
        ```python
        >>> from rattler.package.package_name import PackageName
        >>> from rattler.version.version import Version
        >>> gvp = GenericVirtualPackage(PackageName("__archspec"), Version("1"), "x86_64")
        >>> gvp
        GenericVirtualPackage("__archspec=1=x86_64")
        >>>
        ```
        """
        return f'GenericVirtualPackage("{self._generic_virtual_package.as_str()}")'



from __future__ import annotations
from typing import Dict, List, Optional, Tuple
from rattler.lock.channel import LockChannel
from rattler.lock.package import LockedPackage
from rattler.lock.pypi import PypiPackageData, PypiPackageEnvironmentData
from rattler.platform.platform import Platform

from rattler.rattler import PyEnvironment
from rattler.repo_data.record import RepoDataRecord


class Environment:
    """
    Information about a specific environment in the lock-file.
    """

    _env: PyEnvironment

    def __init__(self, name: str, requirements: Dict[Platform, List[RepoDataRecord]]) -> None:
        self._env = PyEnvironment(
            name,
            # TODO: move this logic to rust
            {platform._inner: [record._record for record in records] for (platform, records) in requirements.items()},
        )

    def platforms(self) -> List[Platform]:
        """
        Returns all the platforms for which we have a locked-down environment.

        Examples
        --------
        ```python
        >>> from rattler import LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> env.platforms()
        [...]
        >>>
        ```
        """
        return [Platform._from_py_platform(p) for p in self._env.platforms()]

    def channels(self) -> List[LockChannel]:
        """
        Returns the channels that are used by this environment.
        Note that the order of the channels is significant.
        The first channel is the highest priority channel.

        Examples
        --------
        ```python
        >>> from rattler import LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> env.channels()
        [LockChannel(url="https://conda.anaconda.org/conda-forge/")]
        >>>
        ```
        """
        return [LockChannel._from_py_lock_channel(c) for c in self._env.channels()]

    def packages(self, platform: Platform) -> Optional[List[LockedPackage]]:
        """
        Returns all the packages for a specific platform in this environment.

        Examples
        --------
        ```python
        >>> from rattler import Platform, LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> env.packages(Platform("osx-arm64"))[0]
        LockedPackage()
        >>>
        ```
        """
        if packages := self._env.packages(platform._inner):
            return [LockedPackage._from_py_lock_package(p) for p in packages]
        return None

    def packages_by_platform(self) -> Dict[Platform, List[LockedPackage]]:
        """
        Returns a list of all packages and platforms defined for this environment.

        Examples
        --------
        ```python
        >>> from rattler import LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pkgs = env.packages_by_platform()
        >>> list(pkgs.keys())
        [Platform(...)]
        >>>
        ```
        """
        return {
            Platform._from_py_platform(platform): [LockedPackage._from_py_lock_package(p) for p in packages]
            for (platform, packages) in self._env.packages_by_platform()
        }

    def pypi_packages(
        self,
    ) -> Dict[Platform, List[Tuple[PypiPackageData, PypiPackageEnvironmentData]]]:
        """
        Returns all pypi packages for all platforms.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> pypi_packages[Platform("osx-arm64")][0]
        (PypiPackageData(), PypiPackageEnvironmentData())
        >>>
        ```
        """
        return {
            Platform._from_py_platform(platform): [
                (
                    PypiPackageData._from_py_pypi_package_data(pkg_data),
                    PypiPackageEnvironmentData._from_py_pypi_env_data(env_data),
                )
                for (pkg_data, env_data) in pypi_tup
            ]
            for (platform, pypi_tup) in self._env.pypi_packages().items()
        }

    def conda_repodata_records(self) -> Dict[Platform, List[RepoDataRecord]]:
        """
        Returns all conda packages for all platforms.

        Examples
        --------
        ```python
        >>> from rattler import LockFile
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> env.conda_repodata_records()
        {'osx-arm64': [RepoDataRecord(...), ...]}
        >>>
        ```
        """
        return {
            platform.name: [RepoDataRecord._from_py_record(r) for r in records]
            for (platform, records) in self._env.conda_repodata_records().items()
        }

    def conda_repodata_records_for_platform(self, platform: Platform) -> Optional[List[RepoDataRecord]]:
        """
        Takes all the conda packages, converts them to [`RepoDataRecord`] and returns them or
        returns an error if the conversion failed. Returns `None` if the specified platform is not
        defined for this environment.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> rdr = env.conda_repodata_records_for_platform(Platform("osx-arm64"))
        >>> rdr
        [...]
        >>> rdr[0]
        RepoDataRecord(...)
        >>>
        ```
        """
        if records := self._env.conda_repodata_records_for_platform(platform._inner):
            return [RepoDataRecord._from_py_record(r) for r in records]
        return None

    def pypi_packages_for_platform(
        self, platform: Platform
    ) -> Optional[List[Tuple[PypiPackageData, PypiPackageEnvironmentData]]]:
        """
        Returns all the pypi packages and their associated environment data for the specified
        platform. Returns `None` if the platform is not defined for this environment.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> osx_pypi_pkgs = env.pypi_packages_for_platform(Platform("osx-arm64"))
        >>> osx_pypi_pkgs
        [...]
        >>> osx_pypi_pkgs[0]
        (PypiPackageData(), PypiPackageEnvironmentData())
        >>>
        ```
        """
        if data := self._env.pypi_packages_for_platform(platform._inner):
            return [
                (
                    PypiPackageData._from_py_pypi_package_data(pkg_data),
                    PypiPackageEnvironmentData._from_py_pypi_env_data(env_data),
                )
                for (pkg_data, env_data) in data
            ]
        return None

    @classmethod
    def _from_py_environment(cls, py_environment: PyEnvironment) -> Environment:
        """
        Construct Rattler Environment from FFI PyEnvironment object.
        """
        env = cls.__new__(cls)
        env._env = py_environment
        return env

    def __repr__(self) -> str:
        """
        Returns a representation of the Environment.
        """
        return "Environment()"



from __future__ import annotations

from rattler.rattler import PyRecord
from rattler.repo_data import PackageRecord


class RepoDataRecord(PackageRecord):
    @property
    def url(self) -> str:
        """
        The canonical URL from where to get this package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.url
        'https://conda.anaconda.org/conda-forge/win-64/libsqlite-3.40.0-hcfcfb64_0.tar.bz2'
        >>>
        ```
        """
        return self._record.url

    @property
    def channel(self) -> str:
        """
        String representation of the channel where the
        package comes from. This could be a URL but it
        could also be a channel name.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.channel
        'https://conda.anaconda.org/conda-forge/win-64'
        >>>
        ```
        """
        return self._record.channel

    @property
    def file_name(self) -> str:
        """
        The filename of the package.

        Examples
        --------
        ```python
        >>> from rattler import PrefixRecord
        >>> record = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/libsqlite-3.40.0-hcfcfb64_0.json"
        ... )
        >>> record.file_name
        'libsqlite-3.40.0-hcfcfb64_0.tar.bz2'
        >>>
        ```
        """
        return self._record.file_name

    @classmethod
    def _from_py_record(cls, py_record: PyRecord) -> RepoDataRecord:
        """
        Construct Rattler RepoDataRecord from FFI PyRecord object.
        """

        # quick sanity check
        assert py_record.is_repodata_record
        record = cls.__new__(cls)
        record._record = py_record
        return record

    def __repr__(self) -> str:
        """
        Returns a representation of the RepoDataRecord.

        Examples
        --------
        ```python
        >>> from rattler import RepoData, Channel
        >>> repo_data = RepoData(
        ...     "../test-data/test-server/repo/noarch/repodata.json"
        ... )
        >>> repo_data.into_repo_data(Channel("test"))[0]
        RepoDataRecord(url="https://conda.anaconda.org/test/noarch/test-package-0.1-0.tar.bz2")
        >>>
        ```
        """
        return f'{type(self).__name__}(url="{self.url}")'



from __future__ import annotations

from rattler.rattler import PyLockChannel


class LockChannel:
    _channel: PyLockChannel

    def __init__(self, url: str) -> None:
        self._channel = PyLockChannel(url)

    @classmethod
    def _from_py_lock_channel(cls, channel: PyLockChannel) -> LockChannel:
        """
        Construct Rattler LockChannel from FFI PyLockChannel object.
        """
        chan = cls.__new__(cls)
        chan._channel = channel
        return chan

    def __str__(self) -> str:
        """
        Returns a string representation of the LockChannel.

        Examples
        --------
        ```python
        >>> channel = LockChannel("https://conda.anaconda.org/conda-forge/")
        >>> str(channel)
        'https://conda.anaconda.org/conda-forge/'
        >>>
        ```
        """
        return self._channel.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the LockChannel.

        Examples
        --------
        ```python
        >>> channel = LockChannel("https://conda.anaconda.org/conda-forge/")
        >>> channel
        LockChannel(url="https://conda.anaconda.org/conda-forge/")
        >>>
        ```
        """
        return f'LockChannel(url="{self._channel.as_str()}")'



from __future__ import annotations
import os
from pathlib import Path
from typing import List

from rattler.rattler import PyRunExportsJson


class RunExportsJson:
    """
    A representation of the `run_exports.json` file found in package archives.
    The `run_exports.json` file contains information about the run exports of a package
    """

    _inner: PyRunExportsJson

    @staticmethod
    def from_package_archive(path: os.PathLike[str]) -> RunExportsJson:
        """
        Parses the package file from archive.
        Note: If you want to extract multiple `info/*` files then this will be slightly
              slower than manually iterating over the archive entries with
              custom logic as this skips over the rest of the archive

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports
        RunExportsJson()
        >>>
        ```
        """
        return RunExportsJson._from_py_run_exports_json(PyRunExportsJson.from_package_archive(path))

    @staticmethod
    def from_path(path: os.PathLike[str]) -> RunExportsJson:
        """
        Parses the object from a file specified by a `path`, using a format
        appropriate for the file type.

        For example, if the file is in JSON format, this function reads the data
        from the file at the specified path, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not read, this function returns an error.
        """
        return RunExportsJson._from_py_run_exports_json(PyRunExportsJson.from_path(Path(path)))

    @staticmethod
    def from_package_directory(path: os.PathLike[str]) -> RunExportsJson:
        """
        Parses the object by looking up the appropriate file from the root of the
        specified Conda archive directory, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function reads the
        appropriate file from the archive, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not be read, this function returns an error.
        """
        return RunExportsJson._from_py_run_exports_json(PyRunExportsJson.from_package_directory(Path(path)))

    @staticmethod
    def from_str(string: str) -> RunExportsJson:
        """
        Parses the object from a string, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function parses the JSON
        string and returns the resulting object. If the file is not in a parsable
        format, this function returns an error.
        """
        return RunExportsJson._from_py_run_exports_json(PyRunExportsJson.from_str(string))

    @staticmethod
    def package_path() -> str:
        """
        Returns the path to the file within the Conda archive.

        The path is relative to the root of the archive and includes any necessary
        directories.
        """
        return PyRunExportsJson.package_path()

    @property
    def weak(self) -> List[str]:
        """
        Weak run exports apply a dependency from host to run.

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports.weak
        ['python_abi 3.10.* *_cp310']
        >>>
        ```
        """
        return self._inner.weak

    @property
    def strong(self) -> List[str]:
        """
        Strong run exports apply a dependency from build to host and run.

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports.strong
        []
        >>>
        ```
        """
        return self._inner.strong

    @property
    def noarch(self) -> List[str]:
        """
        NoArch run exports apply a run export only to noarch packages (other run exports are ignored).
        For example, python uses this to apply a dependency on python to all noarch packages, but not to
        the python_abi package.

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports.noarch
        ['python']
        >>>
        ```
        """
        return self._inner.noarch

    @property
    def weak_constrains(self) -> List[str]:
        """
        Weak constrains apply a constrain dependency from host to build, or run to host.

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports.weak_constrains
        []
        >>>
        ```
        """
        return self._inner.weak_constrains

    @property
    def strong_constrains(self) -> List[str]:
        """
        Strong constrains apply a constrain dependency from build to host and run.

        Examples
        --------
        ```python
        >>> run_exports = RunExportsJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> run_exports.strong_constrains
        []
        >>>
        ```
        """
        return self._inner.strong_constrains

    @classmethod
    def _from_py_run_exports_json(cls, py_run_exports_json: PyRunExportsJson) -> RunExportsJson:
        run_exports_json = cls.__new__(cls)
        run_exports_json._inner = py_run_exports_json

        return run_exports_json

    def __repr__(self) -> str:
        """
        Returns a representation of the RunExportsJson.
        """
        return "RunExportsJson()"



from __future__ import annotations
import os
from typing import Dict, List, Optional, Tuple
from rattler.lock.environment import Environment

from rattler.rattler import PyLockFile


class LockFile:
    """
    Represents a lock-file for both Conda packages and Pypi packages.
    Lock-files can store information for multiple platforms and for multiple environments.
    """

    _lock_file: PyLockFile

    def __init__(self, envs: Dict[str, Environment]) -> None:
        self._lock_file = PyLockFile({name: env._env for (name, env) in envs.items()})

    @staticmethod
    def from_path(path: os.PathLike[str]) -> LockFile:
        """
        Parses a rattler-lock file from a file.

        Examples
        --------
        ```python
        >>> lock_file = LockFile.from_path("./pixi.lock")
        >>> lock_file
        LockFile()
        >>>
        ```
        """
        return LockFile._from_py_lock_file(PyLockFile.from_path(path))

    def to_path(self, path: os.PathLike[str]) -> None:
        """
        Writes the rattler-lock to a file.

        Examples
        --------
        ```python
        >>> lock_file = LockFile.from_path("./pixi.lock")
        >>> lock_file.to_path("/tmp/test.lock")
        >>>
        ```
        """
        return self._lock_file.to_path(path)

    def environments(self) -> List[Tuple[str, Environment]]:
        """
        Returns an iterator over all environments defined in the lock-file.

        Examples
        --------
        ```python
        >>> lock_file = LockFile.from_path("./pixi.lock")
        >>> lock_file.environments()
        [('default', Environment()), ('docs', Environment()), ('build', Environment()), ('test', Environment())]
        >>>
        ```
        """
        return [(name, Environment._from_py_environment(e)) for (name, e) in self._lock_file.environments()]

    def environment(self, name: str) -> Optional[Environment]:
        """
        Returns the environment with the given name.

        Examples
        --------
        ```python
        >>> lock_file = LockFile.from_path("./pixi.lock")
        >>> lock_file.environment("default")
        Environment()
        >>> lock_file.environment("doesnt-exist")
        >>>
        ```
        """
        if env := self._lock_file.environment(name):
            return Environment._from_py_environment(env)
        return None

    def default_environment(self) -> Optional[Environment]:
        """
        Returns the environment with the default name as defined by [`DEFAULT_ENVIRONMENT_NAME`].

        Examples
        --------
        ```python
        >>> lock_file = LockFile.from_path("./pixi.lock")
        >>> lock_file.default_environment()
        Environment()
        >>>
        ```
        """
        return Environment._from_py_environment(self._lock_file.default_environment())

    @classmethod
    def _from_py_lock_file(cls, py_lock_file: PyLockFile) -> LockFile:
        """
        Construct Rattler LockFile from FFI PyLockFile object.
        """
        lock_file = cls.__new__(cls)
        lock_file._lock_file = py_lock_file
        return lock_file

    def __repr__(self) -> str:
        """
        Returns a representation of the LockFile.
        """
        return "LockFile()"



from __future__ import annotations
import os
from typing import List, Optional

from rattler.rattler import PyRecord
from rattler.prefix.prefix_paths import PrefixPaths
from rattler.repo_data.record import RepoDataRecord


class PrefixRecord(RepoDataRecord):
    @classmethod
    def _from_py_record(cls, py_record: PyRecord) -> PrefixRecord:
        """Construct Rattler PrefixRecord from FFI PyRecord object."""

        # quick sanity check
        assert py_record.is_prefix_record
        record = cls.__new__(cls)
        record._record = py_record
        return record

    @staticmethod
    def from_path(path: os.PathLike[str]) -> PrefixRecord:
        """
        Parses a PrefixRecord from a file.

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> assert isinstance(r, PrefixRecord)
        >>>
        ```
        """
        return PrefixRecord._from_py_record(PyRecord.from_path(path))

    def write_to_path(self, path: os.PathLike[str], pretty: bool) -> None:
        """
        Writes the contents of this instance to the file at the specified location.
        """
        self._record.write_to_path(path, pretty)

    @property
    def package_tarball_full_path(self) -> Optional[os.PathLike[str]]:
        """
        The path to where the archive of the package was stored on disk.

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.package_tarball_full_path
        'C:\\\\Users\\\\bas\\\\micromamba\\\\pkgs\\\\requests-2.28.2-pyhd8ed1ab_0.tar.bz2'
        >>>
        ```
        """
        return self._record.package_tarball_full_path

    @property
    def extracted_package_dir(self) -> Optional[os.PathLike[str]]:
        """
        The path that contains the extracted package content.

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.extracted_package_dir
        'C:\\\\Users\\\\bas\\\\micromamba\\\\pkgs\\\\requests-2.28.2-pyhd8ed1ab_0'
        >>>
        ```
        """
        return self._record.extracted_package_dir

    @property
    def files(self) -> List[os.PathLike[str]]:
        """
        A sorted list of all files included in this package

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.files # doctest:+ELLIPSIS
        [...]
        >>>
        ```
        """
        return self._record.files

    @property
    def paths_data(self) -> PrefixPaths:
        """
        Information about how files have been linked when installing the package.

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.paths_data # doctest:+ELLIPSIS
        PrefixPaths(paths=[...])
        >>>
        ```
        """
        return PrefixPaths._from_py_prefix_paths(self._record.paths_data)

    @property
    def requested_spec(self) -> Optional[str]:
        """
        The spec that was used when this package was installed.
        Note that this field is not currently updated if another
        spec was used. If this package was not directly requested by the
        user but was instead installed as a dependency of another package
        `None` will be returned.

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.requested_spec
        ''
        >>>
        ```
        """
        return self._record.requested_spec

    def __repr__(self) -> str:
        """
        Returns a representation of the version

        Examples
        --------
        ```python
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r
        PrefixRecord(name="requests", version="2.28.2")
        >>>
        ```
        """
        return f'PrefixRecord(name="{self.name.normalized}", version="{self.version}")'



from __future__ import annotations
import datetime
from typing import List, Optional
from rattler.match_spec.match_spec import MatchSpec

from rattler.channel import ChannelPriority
from rattler.rattler import py_solve
from rattler.repo_data.record import RepoDataRecord
from rattler.repo_data.sparse import SparseRepoData
from rattler.virtual_package.generic import GenericVirtualPackage


def solve(
    specs: List[MatchSpec],
    available_packages: List[SparseRepoData],
    locked_packages: Optional[List[RepoDataRecord]] = None,
    pinned_packages: Optional[List[RepoDataRecord]] = None,
    virtual_packages: Optional[List[GenericVirtualPackage]] = None,
    timeout: Optional[datetime.timedelta] = None,
    channel_priority: ChannelPriority = ChannelPriority.Strict,
) -> List[RepoDataRecord]:
    """
    Resolve the dependencies and return the `RepoDataRecord`s
    that should be present in the environment.

    Arguments:
        specs: A list of matchspec to solve.
        available_packages: A list of RepoData to use for solving the `specs`.
        locked_packages: Records of packages that are previously selected.
                         If the solver encounters multiple variants of a single
                         package (identified by its name), it will sort the records
                         and select the best possible version. However, if there
                         exists a locked version it will prefer that variant instead.
                         This is useful to reduce the number of packages that are
                         updated when installing new packages. Usually you add the
                         currently installed packages or packages from a lock-file here.
        pinned_packages: Records of packages that are previously selected and CANNOT
                         be changed. If the solver encounters multiple variants of
                         a single package (identified by its name), it will sort the
                         records and select the best possible version. However, if
                         there is a variant available in the `pinned_packages` field it
                         will always select that version no matter what even if that
                         means other packages have to be downgraded.
        virtual_packages: A list of virtual packages considered active.
        channel_priority: (Default = ChannelPriority.Strict) When `ChannelPriority.Strict`
                         the channel that the package is first found in will be used as
                         the only channel for that package. When `ChannelPriority.Disabled`
                         it will search for every package in every channel.

    Returns:
        Resolved list of `RepoDataRecord`s.
    """

    return [
        RepoDataRecord._from_py_record(solved_package)
        for solved_package in py_solve(
            [spec._match_spec for spec in specs],
            [package._sparse for package in available_packages],
            [package._record for package in locked_packages or []],
            [package._record for package in pinned_packages or []],
            [v_package._generic_virtual_package for v_package in virtual_packages or []],
            channel_priority.value,
            timeout.microseconds if timeout else None,
        )
    ]



from enum import Enum

from rattler.rattler import PyChannelPriority


class ChannelPriority(Enum):
    """
    Defines how priority of channels functions during solves. If strict, the channel that the package is first
    found in will be used as the only channel for that package. If disabled, then packages can be retrieved from
    any channel as package version takes precedence.
    """

    Strict = PyChannelPriority.Strict
    Disabled = PyChannelPriority.Disabled



from __future__ import annotations
from enum import Enum

from typing import Iterable, Optional
from pathlib import Path
import os
import sys
from rattler.platform.platform import Platform, PlatformLiteral

from rattler.rattler import (
    PyActivationVariables,
    PyActivator,
    PyShellEnum,
    PyActivationResult,
)


class PathModificationBehavior(Enum):
    """
    The behavior to use when modifying the PATH environment variable.
    Prepend will add the new path to the beginning of the PATH variable.
    Append will add the new path to the end of the PATH variable.
    Replace will replace the entire PATH variable with the new path.
    """

    Prepend = "prepend"
    Append = "append"
    Replace = "replace"


class ActivationVariables:
    """An object that holds the state of the current environment."""

    def __init__(
        self,
        current_prefix: Optional[os.PathLike[str]] = None,
        current_path: Optional[Iterable[str] | Iterable[os.PathLike[str]]] = sys.path,
        path_modification_behavior: PathModificationBehavior = PathModificationBehavior.Prepend,
    ) -> None:
        """
        Construct a new ActivationVariables object.

        current_prefix: The current activated conda prefix (usually
            `sys.env["CONDA_PREFIX"]`). This prefix is going to be deactivated.
        current_path: The current PATH environment variable (usually `sys.path`).
        path_modification_behavior: The behavior to use when modifying the PATH
            environment variable. One of "Prepend", "Append", or "Replace".
            Defaults to "Prepend".
        """
        self._activation_variables = PyActivationVariables(
            current_prefix, current_path, path_modification_behavior.value
        )

    def __str__(self) -> str:
        return self._activation_variables.as_str()


class ActivationResult:
    """An object that holds the result of activating a conda environment."""

    _py_activation_result: PyActivationResult

    @classmethod
    def _from_py_activation_result(cls, py_activation_result: PyActivationResult) -> ActivationResult:
        """Construct Rattler version from FFI PyActivationResult object."""
        activation_result = cls.__new__(cls)
        activation_result._py_activation_result = py_activation_result
        return activation_result

    @property
    def path(self) -> Path:
        """The new PATH environment variable."""
        return self._py_activation_result.path

    @property
    def script(self) -> str:
        """The script to run to activate the environment."""
        return self._py_activation_result.script


class Shell:
    """An enum of supported shells."""

    bash = PyShellEnum.Bash
    zsh = PyShellEnum.Zsh
    fish = PyShellEnum.Fish
    xonsh = PyShellEnum.Xonsh
    powershell = PyShellEnum.PowerShell
    cmd_exe = PyShellEnum.CmdExe


def activate(
    prefix: Path,
    activation_variables: ActivationVariables,
    shell: Optional[Shell],
    platform: Optional[Platform | PlatformLiteral] = None,
) -> ActivationResult:
    """
    Return an ActivationResult object that contains the new PATH environment variable
    and the script to run to activate the environment.

    Arguments:
        prefix: The path to the conda prefix to activate.
        activation_variables: The current activation variables.
        shell: The shell to generate the activation script for.
        platform: The platform to generate the activation script for.
                  If None, the current platform is used.

    Returns:
        An ActivationResult object containing the new PATH environment variable
        and the script to run to activate the environment.

    Examples
    --------
    ```python
    >>> from rattler.shell import Shell, activate, ActivationVariables
    >>> from rattler.platform import Platform
    >>> from pathlib import Path
    >>> import sys
    >>> p = Path("/path/to/conda/prefix")
    >>> actvars = ActivationVariables(None, sys.path)
    >>> a = activate(p, actvars, Shell.xonsh)
    >>> print(a)
    <rattler.shell.shell.ActivationResult object at ...>
    >>>
    ```
    """
    platform = Platform(platform) if isinstance(platform, str) else platform or Platform.current()
    shell = shell or Shell.bash
    return ActivationResult._from_py_activation_result(
        PyActivator.activate(prefix, activation_variables._activation_variables, platform._inner, shell)
    )



try:
    from rattler.rattler import (
        InvalidVersionError,
        InvalidMatchSpecError,
        InvalidPackageNameError,
        InvalidUrlError,
        InvalidChannelError,
        ActivationError,
        ParsePlatformError,
        ParseArchError,
        TransactionError,
        LinkError,
        IoError,
        DetectVirtualPackageError,
        CacheDirError,
        FetchRepoDataError,
        SolverError,
        ConvertSubdirError,
        VersionBumpError,
        EnvironmentCreationError,
        ExtractError,
    )
except ImportError:
    # They are only redefined for documentation purposes
    # when there is no binary yet

    class InvalidVersionError(Exception):  # type: ignore[no-redef]
        """Error that can occur when parsing a Version"""

    class InvalidMatchSpecError(Exception):  # type: ignore[no-redef]
        """Error that can occur when parsing a MatchSpec"""

    class InvalidPackageNameError(Exception):  # type: ignore[no-redef]
        """Error that can occur when parsing a package name"""

    class InvalidUrlError(Exception):  # type: ignore[no-redef]
        """Error that can occur when parsing a URL"""

    class InvalidChannelError(Exception):  # type: ignore[no-redef]
        """Error that can occur when parsing a channel."""

    class ActivationError(Exception):  # type: ignore[no-redef]
        """Error that can occur when activating a conda environment"""

    class ParsePlatformError(Exception):  # type: ignore[no-redef]
        """An error that can occur when parsing a platform from a string."""

    class ParseArchError(Exception):  # type: ignore[no-redef]
        """An error that can occur when parsing an arch from a string."""

    class TransactionError(Exception):  # type: ignore[no-redef]
        """An error that can occur when executing a transaction"""

    class LinkError(Exception):  # type: ignore[no-redef]
        """An error that can occur when linking a package"""

    class IoError(Exception):  # type: ignore[no-redef]
        """An error that can occur during io operations"""

    class DetectVirtualPackageError(Exception):  # type: ignore[no-redef]
        """An error that can occur when trying to detect virtual packages"""

    class CacheDirError(Exception):  # type: ignore[no-redef]
        """An error that can occur when querying the cache directory"""

    class FetchRepoDataError(Exception):  # type: ignore[no-redef]
        """An error that can occur when fetching repo data"""

    class SolverError(Exception):  # type: ignore[no-redef]
        """An error that can occur when trying to solve an environment"""

    class ConvertSubdirError(Exception):  # type: ignore[no-redef]
        """An error that can occur when parsing a platform from a string."""

    class VersionBumpError(Exception):  # type: ignore[no-redef]
        """An error that can occur when bumping a version."""

    class EnvironmentCreationError(Exception):  # type: ignore[no-redef]
        """An error that can occur when creating an environment."""

    class ExtractError(Exception):  # type: ignore[no-redef]
        """An error that can occur when extracting an archive."""


__all__ = [
    "ActivationError",
    "CacheDirError",
    "DetectVirtualPackageError",
    "FetchRepoDataError",
    "InvalidChannelError",
    "InvalidMatchSpecError",
    "InvalidPackageNameError",
    "InvalidUrlError",
    "InvalidVersionError",
    "IoError",
    "LinkError",
    "ParseArchError",
    "ParsePlatformError",
    "SolverError",
    "TransactionError",
    "ConvertSubdirError",
    "VersionBumpError",
    "EnvironmentCreationError",
    "ExtractError",
]



from __future__ import annotations

from rattler.rattler import PyPatchInstructions


class PatchInstructions:
    _patch_instructions: PyPatchInstructions

    @classmethod
    def _from_py_patch_instructions(cls, py_patch_instructions: PyPatchInstructions) -> PatchInstructions:
        """
        Construct Rattler PatchInstructions from FFI PyPatchInstructions object.
        """
        patch_instructions = cls.__new__(cls)
        patch_instructions._patch_instructions = py_patch_instructions
        return patch_instructions

    def __repr__(self) -> str:
        """
        Returns a representation of the PatchInstructions.
        """
        return f"{type(self).__name__}()"



# type: ignore
import os
from pathlib import Path
import pytest
import shutil

from rattler import Platform, index


@pytest.fixture
def package_directory(tmp_path) -> Path:
    data_dir = Path(os.path.join(os.path.dirname(__file__), "../../../test-data/"))

    win_filename = "ruff-0.0.171-py310h298983d_0.conda"
    noarch_filename = "pytweening-1.0.4-pyhd8ed1ab_0.tar.bz2"
    win_subdir = tmp_path / "win-64"
    noarch_subdir = tmp_path / "noarch"
    win_subdir.mkdir()
    noarch_subdir.mkdir()
    shutil.copy(data_dir / win_filename, win_subdir / win_filename)
    shutil.copy(data_dir / noarch_filename, noarch_subdir / noarch_filename)
    return tmp_path


def test_index(package_directory):
    index(package_directory)

    assert set(os.listdir(package_directory)) == {"noarch", "win-64"}
    assert "repodata.json" in os.listdir(package_directory / "win-64")
    with open(package_directory / "win-64/repodata.json") as f:
        assert "ruff-0.0.171-py310h298983d_0" in f.read()
    assert "repodata.json" in os.listdir(package_directory / "noarch")
    with open(package_directory / "noarch/repodata.json") as f:
        assert "pytweening-1.0.4-pyhd8ed1ab_0" in f.read()


def test_index_specific_subdir_non_noarch(package_directory):
    index(package_directory, Platform("win-64"))

    assert "repodata.json" in os.listdir(package_directory / "win-64")
    with open(package_directory / "win-64/repodata.json") as f:
        assert "ruff-0.0.171-py310h298983d_0" in f.read()
    assert "repodata.json" in os.listdir(package_directory / "noarch")
    with open(package_directory / "noarch/repodata.json") as f:
        assert "pytweening-1.0.4-pyhd8ed1ab_0" in f.read()


def test_index_specific_subdir_noarch(package_directory):
    index(package_directory, Platform("noarch"))

    win_files = os.listdir(package_directory / "win-64")
    assert "repodata.json" not in win_files
    assert "ruff-0.0.171-py310h298983d_0.conda" in win_files
    assert "repodata.json" in os.listdir(package_directory / "noarch")
    with open(package_directory / "noarch/repodata.json") as f:
        assert "pytweening-1.0.4-pyhd8ed1ab_0" in f.read()



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from rattler.channel.channel import Channel

from rattler.rattler import PyMatchSpec

from rattler.package.package_name import PackageName

if TYPE_CHECKING:
    from rattler.match_spec import NamelessMatchSpec
    from rattler.repo_data import PackageRecord


class MatchSpec:
    """
    A `MatchSpec` is a query language for conda packages.
    It can be composed of any of the attributes of `PackageRecord`.

    `MatchSpec` can be composed of keyword arguments, where keys are
    any of the attributes of `PackageRecord`. Values for keyword arguments
    are exact values the attributes should match against. Many fields can
    be matched against non-exact values by including wildcard `*` and `>`/`<`
    ranges where supported. Any non-specified field is the equivalent of a
    full wildcard match.

    MatchSpecs can also be composed using a single positional argument, with optional
    keyword arguments. Keyword arguments also override any conflicting information
    provided in the positional argument. Conda has historically had several string
    representations for equivalent MatchSpecs.

    A series of rules are now followed for creating the canonical string
    representation of a MatchSpec instance. The canonical string representation can
    generically be represented by:

    `(channel(/subdir):(namespace):)name(version(build))[key1=value1,key2=value2]`

    where `()` indicate optional fields.

    The rules for constructing a canonical string representation are:

    1. `name` (i.e. "package name") is required, but its value can be '*'. Its
    position is always outside the key-value brackets.
    2. If `version` is an exact version, it goes outside the key-value brackets and
    is prepended by `==`. If `version` is a "fuzzy" value (e.g. `1.11.*`), it goes
    outside the key-value brackets with the `.*` left off and is prepended by `=`.
    Otherwise `version` is included inside key-value brackets.
    3. If `version` is an exact version, and `build` is an exact value, `build` goes
    outside key-value brackets prepended by a `=`.  Otherwise, `build` goes inside
    key-value brackets. `build_string` is an alias for `build`.
    4. The `namespace` position is being held for a future feature. It is currently
    ignored.
    5. If `channel` is included and is an exact value, a `::` separator is used between
    `channel` and `name`.  `channel` can either be a canonical channel name or a
    channel url. In the canonical string representation, the canonical channel name
    will always be used.
    6. If `channel` is an exact value and `subdir` is an exact value, `subdir` is
    appended to `channel` with a `/` separator.  Otherwise, `subdir` is included in
    the key-value brackets.
    7. Key-value brackets can be delimited by comma, space, or comma+space. Value can
    optionally be wrapped in single or double quotes, but must be wrapped if `value`
    contains a comma, space, or equal sign.  The canonical format uses comma delimiters
    and single quotes.
    8. When constructing a `MatchSpec` instance from a string, any key-value pair given
    inside the key-value brackets overrides any matching parameter given outside the
    brackets.

    When `MatchSpec` attribute values are simple strings, the are interpreted using the
    following conventions:
    - If the string begins with `^` and ends with `$`, it is converted to a regex.
    - If the string contains an asterisk (`*`), it is transformed from a glob to a
    regex.
    - Otherwise, an exact match to the string is sought.

    To fully-specify a package with a full, exact spec, the following fields must be
    given as exact values:
    - channel
    - subdir
    - name
    - version
    - build
    """

    def __init__(self, spec: str, strict: bool = False) -> None:
        if isinstance(spec, str):
            self._match_spec = PyMatchSpec(spec, strict)
        else:
            raise TypeError(
                "MatchSpec constructor received unsupported type" f" {type(spec).__name__!r} for the 'spec' parameter"
            )

    @property
    def name(self) -> Optional[PackageName]:
        """
        The name of the package.
        """
        return PackageName._from_py_package_name(self._match_spec.name)

    @property
    def version(self) -> Optional[str]:
        """
        The version spec of the package (e.g. `1.2.3`, `>=1.2.3`, `1.2.*`)
        """
        return self._match_spec.version

    @property
    def build(self) -> Optional[str]:
        """
        The build string of the package (e.g. `py37_0`, `py37h6de7cb9_0`, `py*`)
        """
        return self._match_spec.build

    @property
    def build_number(self) -> Optional[str]:
        """
        The build number of the package.
        """
        return self._match_spec.build_number

    @property
    def file_name(self) -> Optional[str]:
        """
        Match the specific filename of the package.
        """
        return self._match_spec.file_name

    @property
    def channel(self) -> Optional[Channel]:
        """
        The channel of the package.
        """
        if (channel := self._match_spec.channel) is not None:
            return Channel(channel.name)
        return None

    @property
    def subdir(self) -> Optional[str]:
        """
        The subdir of the channel.
        """
        return self._match_spec.subdir

    @property
    def namespace(self) -> Optional[str]:
        """
        The namespace of the package.
        """
        return self._match_spec.namespace

    @property
    def md5(self) -> Optional[bytes]:
        """
        The md5 hash of the package.
        """
        return self._match_spec.md5

    @property
    def sha256(self) -> Optional[bytes]:
        """
        The sha256 hash of the package.
        """
        return self._match_spec.sha256

    @classmethod
    def _from_py_match_spec(cls, py_match_spec: PyMatchSpec) -> MatchSpec:
        """
        Construct py-rattler MatchSpec from PyMatchSpec FFI object.
        """
        match_spec = cls.__new__(cls)
        match_spec._match_spec = py_match_spec

        return match_spec

    def matches(self, record: PackageRecord) -> bool:
        """Match a MatchSpec against a PackageRecord."""
        return self._match_spec.matches(record._record)

    @classmethod
    def from_nameless(cls, spec: NamelessMatchSpec, name: str) -> MatchSpec:
        """
        Constructs a MatchSpec from a NamelessMatchSpec
        and a name.

        Examples
        --------
        ```python
        >>> from rattler import NamelessMatchSpec
        >>> spec = NamelessMatchSpec('3.4')
        >>> MatchSpec.from_nameless(spec, "foo")
        MatchSpec("foo ==3.4")
        >>> MatchSpec.from_nameless(spec, "$foo") # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        exceptions.InvalidPackageNameException
        >>>
        ```
        """
        return cls._from_py_match_spec(PyMatchSpec.from_nameless(spec._nameless_match_spec, name))

    def __str__(self) -> str:
        """
        Returns a string representation of the MatchSpec.

        Examples
        --------
        ```python
        >>> from rattler import NamelessMatchSpec
        >>> spec = NamelessMatchSpec('3.4')
        >>> str(MatchSpec.from_nameless(spec, "foo"))
        'foo ==3.4'
        >>>
        ```
        """
        return self._match_spec.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the MatchSpec.

        Examples
        --------
        ```python
        >>> from rattler import NamelessMatchSpec
        >>> spec = NamelessMatchSpec('3.4')
        >>> MatchSpec.from_nameless(spec, "foo")
        MatchSpec("foo ==3.4")
        >>>
        ```
        """
        return f'MatchSpec("{self._match_spec.as_str()}")'



from rattler.prefix.prefix_record import PrefixRecord
from rattler.prefix.prefix_paths import PrefixPaths, PrefixPathsEntry, PrefixPathType

__all__ = ["PrefixRecord", "PrefixPaths", "PrefixPathsEntry", "PrefixPathType"]



from __future__ import annotations

from rattler.rattler import PyArch

from typing import Literal

ArchLiteral = Literal[
    "x86",
    "x86_64",
    "aarch64",
    "armv6l",
    "armv7l",
    "ppc64le",
    "ppc64",
    "s390x",
    "riscv32",
    "riscv64",
]


class Arch:
    def __init__(self, value: ArchLiteral) -> None:
        self._inner = PyArch(value)

    @classmethod
    def _from_py_arch(cls, py_arch: PyArch) -> Arch:
        """Construct Rattler version from FFI PyArch object."""
        arch = cls.__new__(cls)
        arch._inner = py_arch
        return arch

    def __str__(self) -> str:
        """
        Returns a string representation of the architecture.

        Examples
        --------
        ```python
        >>> str(Arch("x86_64"))
        'x86_64'
        >>>
        ```
        """
        return self._inner.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the architecture.

        Examples
        --------
        ```python
        >>> Arch("aarch64")
        Arch(aarch64)
        >>>
        ```
        """
        return f"Arch({self._inner.as_str()})"



from rattler.channel.channel import Channel
from rattler.channel.channel_config import ChannelConfig
from rattler.channel.channel_priority import ChannelPriority

__all__ = ["Channel", "ChannelConfig", "ChannelPriority"]



from __future__ import annotations
import os
from pathlib import Path
from typing import List, Optional

from rattler.package.package_name import PackageName
from rattler.rattler import PyIndexJson


class IndexJson:
    _inner: PyIndexJson

    @staticmethod
    def from_package_archive(path: os.PathLike[str]) -> IndexJson:
        """
        Parses the package file from archive.
        Note: If you want to extract multiple `info/*` files then this will be slightly
              slower than manually iterating over the archive entries with
              custom logic as this skips over the rest of the archive

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json
        IndexJson()
        >>>
        ```
        """
        return IndexJson._from_py_index_json(PyIndexJson.from_package_archive(path))

    @staticmethod
    def from_package_directory(path: os.PathLike[str]) -> IndexJson:
        """
        Parses the object by looking up the appropriate file from the root of the
        specified Conda archive directory, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function reads the
        appropriate file from the archive, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not be read, this function returns an error.
        """
        return IndexJson._from_py_index_json(PyIndexJson.from_package_directory(Path(path)))

    @staticmethod
    def from_str(string: str) -> IndexJson:
        """
        Parses the object from a string, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function parses the JSON
        string and returns the resulting object. If the file is not in a parsable
        format, this function returns an error.
        """
        return IndexJson._from_py_index_json(PyIndexJson.from_str(string))

    @staticmethod
    def package_path() -> str:
        """
        Returns the path to the file within the Conda archive.

        The path is relative to the root of the archive and includes any necessary
        directories.
        """
        return PyIndexJson.package_path()

    @property
    def arch(self) -> Optional[str]:
        """
        Optionally, the architecture the package is build for.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.arch
        'ppc64le'
        >>>
        ```
        """
        if arch := self._inner.arch:
            return arch

        return None

    @property
    def build(self) -> str:
        """
        The build string of the package.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.build
        'h2c4edbf_0_cpython'
        >>>
        ```
        """
        return self._inner.build

    @property
    def build_number(self) -> int:
        """
        The build number of the package.
        This is also included in the build string.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.build_number
        0
        >>>
        ```
        """
        return self._inner.build_number

    @property
    def constrains(self) -> List[str]:
        """
        The package constraints of the package.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.constrains
        ['python_abi 3.10.* *_cp310']
        >>>
        ```
        """
        return self._inner.constrains

    @property
    def depends(self) -> List[str]:
        """
        The dependencies of the package.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.depends
        ['bzip2 >=1.0.8,<2.0a0', ..., 'xz >=5.2.6,<5.3.0a0']
        >>>
        ```
        """
        return self._inner.depends

    @property
    def features(self) -> Optional[str]:
        """
        Features are a deprecated way to specify different feature sets for the conda solver. This is not
        supported anymore and should not be used. Instead, `mutex` packages should be used to specify
        mutually exclusive features.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.features
        >>>
        ```
        """
        if features := self._inner.features:
            return features

        return None

    @property
    def license(self) -> Optional[str]:
        """
        Optionally, the license.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.license
        'Python-2.0'
        >>>
        ```
        """
        if license := self._inner.license:
            return license

        return None

    @property
    def license_family(self) -> Optional[str]:
        """
        Optionally, the license.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.license_family
        >>>
        ```
        """
        if license_family := self._inner.license_family:
            return license_family

        return None

    @property
    def name(self) -> PackageName:
        """
        The lowercase name of the package.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.name
        PackageName("python")
        >>>
        ```
        """
        return PackageName._from_py_package_name(self._inner.name)

    @property
    def platform(self) -> Optional[str]:
        """
        Optionally, the OS the package is build for.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.platform
        'linux'
        >>>
        ```
        """
        if platform := self._inner.platform:
            return platform

        return None

    @property
    def subdir(self) -> Optional[str]:
        """
        The subdirectory that contains this package.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.subdir
        'linux-ppc64le'
        >>>
        ```
        """
        if subdir := self._inner.subdir:
            return subdir

        return None

    @property
    def timestamp(self) -> Optional[int]:
        """
        The timestamp when this package was created

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.timestamp
        1661200742
        >>>
        ```
        """
        if timestamp := self._inner.timestamp:
            return timestamp

        return None

    @property
    def track_features(self) -> List[str]:
        """
        Track features are nowadays only used to downweight packages (ie. give them less priority). To
        that effect, the number of track features is counted (number of commas) and the package is downweighted
        by the number of track_features.

        Examples
        --------
        ```python
        >>> idx_json = IndexJson.from_package_archive(
        ...     "../test-data/with-symlinks/python-3.10.6-h2c4edbf_0_cpython.tar.bz2"
        ... )
        >>> idx_json.track_features
        []
        >>>
        ```
        """
        return self._inner.track_features

    @classmethod
    def _from_py_index_json(cls, py_index_json: PyIndexJson) -> IndexJson:
        index_json = cls.__new__(cls)
        index_json._inner = py_index_json

        return index_json

    def __repr__(self) -> str:
        """
        Returns a representation of the IndexJson.
        """
        return "IndexJson()"



from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional

from rattler.rattler import PyAboutJson


class AboutJson:
    """
    The `about.json` file contains metadata about the package.
    """

    _inner: PyAboutJson

    @staticmethod
    def from_path(path: os.PathLike[str]) -> AboutJson:
        """
        Parses the object from a file specified by a `path`, using a format
        appropriate for the file type.

        For example, if the file is in JSON format, this function reads the data
        from the file at the specified path, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not read, this function returns an error.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about
        AboutJson()
        >>>
        ```
        """
        return AboutJson._from_py_about_json(PyAboutJson.from_path(Path(path)))

    @staticmethod
    def from_package_directory(path: os.PathLike[str]) -> AboutJson:
        """
        Parses the object by looking up the appropriate file from the root of the
        specified Conda archive directory, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function reads the
        appropriate file from the archive, parse the JSON string and return the
        resulting object. If the file is not in a parsable format or if the file
        could not be read, this function returns an error.
        """
        return AboutJson._from_py_about_json(PyAboutJson.from_package_directory(Path(path)))

    @staticmethod
    def from_str(string: str) -> AboutJson:
        """
        Parses the object from a string, using a format appropriate for the file
        type.

        For example, if the file is in JSON format, this function parses the JSON
        string and returns the resulting object. If the file is not in a parsable
        format, this function returns an error.

        Examples
        --------
        ```python
        >>> import json
        >>> with open("../test-data/dummy-about.json", 'r') as file:
        ...     json_str = json.dumps(json.load(file))
        >>> about = AboutJson.from_str(json_str)
        >>> about
        AboutJson()
        >>>
        ```
        """
        return AboutJson._from_py_about_json(PyAboutJson.from_str(string))

    @staticmethod
    def package_path() -> str:
        """
        Returns the path to the file within the Conda archive.

        The path is relative to the root of the archive and includes any necessary
        directories.

        Examples
        --------
        ```python
        >>> AboutJson.package_path()
        'info/about.json'
        >>>
        ```
        """
        return PyAboutJson.package_path()

    @property
    def channels(self) -> List[str]:
        """
        A list of channels that where used during the build.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.channels
        ['https://conda.anaconda.org/conda-forge']
        >>>
        ```
        """
        return self._inner.channels

    @property
    def description(self) -> Optional[str]:
        """
        Description of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.description
        'A dummy description.'
        >>>
        ```
        """
        if description := self._inner.description:
            return description

        return None

    @property
    def dev_url(self) -> List[str]:
        """
        A list of URLs to the development page of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.dev_url
        ['https://github.com/mamba-org/rattler']
        >>>
        ```
        """
        return self._inner.dev_url

    @property
    def doc_url(self) -> List[str]:
        """
        A list of URLs to the documentation of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.doc_url
        ['https://mamba-org.github.io/rattler/py-rattler/']
        >>>
        ```
        """
        return self._inner.doc_url

    @property
    def home(self) -> List[str]:
        """
        A list URL to the homepage of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.home
        ['http://github.com/mamba-org/rattler']
        >>>
        ```
        """
        return self._inner.home

    @property
    def license(self) -> Optional[str]:
        """
        The license of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.license
        'BSD-3-Clause'
        >>>
        ```
        """
        if license := self._inner.license:
            return license

        return None

    @property
    def license_family(self) -> Optional[str]:
        """
        The license family of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.license_family
        >>> type(about.license_family)
        <class 'NoneType'>
        >>>
        ```
        """
        if license_family := self._inner.license_family:
            return license_family

        return None

    @property
    def source_url(self) -> Optional[str]:
        """
        The URL to the latest source code of the package.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.source_url
        'https://github.com/mamba-org/rattler'
        >>>
        ```
        """
        if source_url := self._inner.source_url:
            return source_url

        return None

    @property
    def summary(self) -> Optional[str]:
        """
        A Short summary description.

        Examples
        --------
        ```python
        >>> about = AboutJson.from_path("../test-data/dummy-about.json")
        >>> about.summary
        'A dummy summary.'
        >>>
        ```
        """
        if summary := self._inner.summary:
            return summary

        return None

    @classmethod
    def _from_py_about_json(cls, py_about_json: PyAboutJson) -> AboutJson:
        about_json = cls.__new__(cls)
        about_json._inner = py_about_json

        return about_json

    def __repr__(self) -> str:
        """
        Returns a representation of the AboutJson.
        """
        return "AboutJson()"



try:
    from rattler.rattler import get_rattler_version as _get_rattler_version

    rattler_version_string = _get_rattler_version()
except ImportError:
    # this is only useful for documentation
    import warnings

    warnings.warn("rattler binary missing!", stacklevel=2)
    rattler_version_string = ""


def get_rattler_version() -> str:
    """
    Return the version of the Python Rattler package as a string.

    If the Rattler binary is missing, returns an empty string.
    """
    return rattler_version_string



from __future__ import annotations
import os
from typing import List, TYPE_CHECKING

from rattler.package.paths_json import FileMode
from rattler.rattler import PyPrefixPaths, PyPrefixPathsEntry, PyPrefixPathType

# `os.PathLike` started to be generic from Python 3.9
if TYPE_CHECKING:
    BasePathLike = os.PathLike[str]
else:
    BasePathLike = os.PathLike


class PrefixPathType:
    _inner: PyPrefixPathType

    @classmethod
    def from_py_path_type(cls, py_path_type: PyPrefixPathType) -> PrefixPathType:
        """Construct Rattler PathType from FFI PyPathType object."""
        path_type = cls.__new__(cls)
        path_type._inner = py_path_type
        return path_type

    @property
    def hardlink(self) -> bool:
        """
        Whether the path should be hardlinked (the default) (once installed)
        """
        return self._inner.hardlink

    @property
    def softlink(self) -> bool:
        """
        Whether the path should be softlinked (once installed)
        """
        return self._inner.softlink

    @property
    def directory(self) -> bool:
        """
        This is a directory
        """
        return self._inner.directory

    @property
    def pyc_file(self) -> bool:
        """
        This is a file compiled from Python code when a noarch package was installed
        """
        return self._inner.pyc_file

    @property
    def windows_python_entry_point_script(self) -> bool:
        """
        A Windows entry point python script (a <entrypoint>-script.py Python script file)
        """
        return self._inner.windows_python_entry_point_script

    @property
    def windows_python_entry_point_exe(self) -> bool:
        """
        A Windows entry point python script (a <entrypoint>.exe executable)
        """
        return self._inner.windows_python_entry_point_exe

    @property
    def unix_python_entrypoint(self) -> bool:
        """
        A Unix entry point python script (a <entrypoint> Python script file)
        """
        return self._inner.unix_python_entrypoint


class PrefixPathsEntry(BasePathLike):
    _inner: PyPrefixPathsEntry

    def __fspath__(self) -> str:
        return str(self._inner.path)

    @classmethod
    def _from_py_paths_entry(cls, py_paths_entry: PyPrefixPathsEntry) -> PrefixPathsEntry:
        """Construct Rattler PathsEntry from FFI PyPathsEntry object."""
        entry = cls.__new__(cls)
        entry._inner = py_paths_entry
        return entry

    @property
    def relative_path(self) -> os.PathLike[str]:
        """
        The relative path of the entry.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/tk-8.6.12-h8ffe710_0.json"
        ... )
        >>> paths = r.paths_data
        >>> relative_path = paths.paths[0].relative_path
        >>>
        ...
        ```
        """
        return self._inner.relative_path

    @property
    def no_link(self) -> bool:
        """
        Whether this file should be linked

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> no_link = paths.paths[0].no_link
        >>>
        ```
        """

    @property
    def path_type(self) -> PrefixPathType:
        """
        The type of the path.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> path_type = paths.paths[0].path_type
        >>>
        ```
        """
        return PrefixPathType.from_py_path_type(self._inner.path_type)

    @property
    def prefix_placeholder(self) -> str | None:
        """
        The prefix placeholder for the path.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> prefix_placeholder = paths.paths[0].prefix_placeholder
        >>>
        ```
        """
        return self._inner.prefix_placeholder

    @property
    def file_mode(self) -> FileMode:
        """
        The file mode of the path.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> file_mode = paths.paths[0].file_mode
        >>>
        ```
        """
        return FileMode._from_py_file_mode(self._inner.file_mode)

    def sha256(self) -> bytes:
        """
        The sha256 of the path.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> sha256 = paths.paths[0].sha256
        >>>
        ```
        """
        return self._inner.sha256

    def sha256_in_prefix(self) -> bytes:
        """
        The sha256 of the path in the prefix.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> sha256_in_prefix = paths.paths[0].sha256_in_prefix
        >>>
        ```
        """
        return self._inner.sha256_in_prefix

    @property
    def size_in_bytes(self) -> int:
        """
        The size of the path in bytes.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> size_in_bytes = paths.paths[0].size_in_bytes
        >>>
        ```
        """
        return self._inner.size_in_bytes


class PrefixPaths:
    _paths: PyPrefixPaths

    @classmethod
    def _from_py_prefix_paths(cls, py_prefix_paths: PyPrefixPaths) -> PrefixPaths:
        """Construct Rattler PrefixRecord from FFI PyPrefixRecord object."""
        paths = cls.__new__(cls)
        paths._paths = py_prefix_paths
        return paths

    @property
    def paths_version(self) -> int:
        """
        The version of the file.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> paths.paths_version
        1
        >>>
        ```
        """
        return self._paths.paths_version

    @property
    def paths(self) -> List[PrefixPathsEntry]:
        """
        All entries included in the package.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> paths.paths # doctest:+ELLIPSIS
        [...]
        >>>
        ```
        """
        return [PrefixPathsEntry._from_py_paths_entry(path) for path in self._paths.paths]

    def __repr__(self) -> str:
        """
        Returns a representation of the PrefixPaths.

        Examples
        --------
        ```python
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.paths_data # doctest:+ELLIPSIS
        PrefixPaths(paths=[...])
        >>>
        ```
        """
        return f"PrefixPaths(paths={self.paths})"



from __future__ import annotations
import os
from rattler.rattler import PyChannelConfig


class ChannelConfig:
    def __init__(self, channel_alias: str = "https://conda.anaconda.org/", root_dir: str = os.getcwd()) -> None:
        """
        Create a new channel configuration.

        Examples
        --------
        ```python
        >>> channel_config = ChannelConfig()
        >>> channel_config # doctest: +ELLIPSIS
        ChannelConfig(channel_alias="https://conda.anaconda.org/", ...
        >>> channel_config = ChannelConfig("https://repo.prefix.dev/", "/path/to/root/dir")
        >>> channel_config
        ChannelConfig(channel_alias="https://repo.prefix.dev/", root_dir="/path/to/root/dir")
        >>>
        ```
        """
        if root_dir is None:
            # Use the current working directory as the root directory.
            root_dir = os.getcwd()
        self._channel_configuration = PyChannelConfig(channel_alias, root_dir)

    def __repr__(self) -> str:
        """
        Return a string representation of this channel configuration.

        Examples
        --------
        ```python
        >>> channel_config = ChannelConfig()
        >>> channel_config # doctest: +ELLIPSIS
        ChannelConfig(channel_alias="https://conda.anaconda.org/", root_dir="...
        >>>
        ```
        """
        alias = self._channel_configuration.channel_alias
        root_dir = self._channel_configuration.root_dir
        return f'ChannelConfig(channel_alias="{alias}", root_dir="{root_dir}")'



# type: ignore
import os
import pytest

from rattler import Channel, SparseRepoData, MatchSpec, solve, link


@pytest.mark.asyncio
async def test_link(tmp_path):
    cache_dir = tmp_path / "cache"
    env_dir = tmp_path / "env"

    linux64_chan = Channel("conda-forge")
    data_dir = os.path.join(os.path.dirname(__file__), "../../../test-data/")
    linux64_path = os.path.join(data_dir, "channels/conda-forge/linux-64/repodata.json")
    linux64_data = SparseRepoData(
        channel=linux64_chan,
        subdir="linux-64",
        path=linux64_path,
    )

    solved_data = solve(
        [MatchSpec("xtensor")],
        [linux64_data],
    )

    await link(solved_data, env_dir, cache_dir)

    assert os.path.exists(env_dir / "include/xtensor.hpp")
    assert os.path.exists(env_dir / "include/xtensor")
    assert os.path.exists(env_dir / "include/xtl")



from __future__ import annotations
import os
from typing import Optional

from rattler.platform import Platform
from rattler.rattler import py_index


def index(
    channel_directory: os.PathLike[str],
    target_platform: Optional[Platform] = None,
) -> None:
    """
    Indexes dependencies in the `channel_directory` for one or more subdirectories within said directory.
    Will generate repodata.json files in each subdirectory containing metadata about each present package,
    or if `target_platform` is specified will only consider the subdirectory corresponding to this platform.
    Will always index the "noarch" subdirectory, and thus this subdirectory should always be present, because
    conda channels at a minimum must include this subdirectory.

    Arguments:
        channel_directory: A `os.PathLike[str]` that is the directory containing subdirectories
                           of dependencies to index.
        target_platform(optional): A `Platform` to index dependencies for.
    """
    py_index(
        channel_directory,
        target_platform._inner if target_platform else target_platform,
    )



from __future__ import annotations
import os
from typing import List, Optional

from rattler.networking.authenticated_client import AuthenticatedClient
from rattler.platform.platform import Platform
from rattler.prefix.prefix_record import PrefixRecord
from rattler.repo_data.record import RepoDataRecord

from rattler.rattler import py_link


async def link(
    dependencies: List[RepoDataRecord],
    target_prefix: os.PathLike[str],
    cache_dir: os.PathLike[str],
    installed_packages: Optional[List[PrefixRecord]] = None,
    platform: Optional[Platform] = None,
    execute_link_scripts: bool = True,
) -> None:
    """
    Create an environment by downloading and linking the `dependencies` in
    the `target_prefix` directory.

    Arguments:
        dependencies: A list of solved `RepoDataRecord`s.
        target_prefix: Path to the directory where the environment should
                       be created.
        cache_dir: Path to directory where the dependencies will be
                   downloaded and cached.
        installed_packages: A list of `PrefixRecord`s which are
                                      already installed in the
                                      `target_prefix`. This can be obtained
                                      by loading `PrefixRecord`s from
                                      `{target_prefix}/conda-meta/`.
        platform: Target platform to create and link the
                            environment. Defaults to current platform.
        execute_link_scripts: Wether to execute the post-link and pre-unlink scripts
                              that may be part of a package. Defaults to True.
    """
    platform = platform or Platform.current()
    client = AuthenticatedClient()

    await py_link(
        dependencies,
        target_prefix,
        cache_dir,
        installed_packages or [],
        platform._inner,
        client._client,
        execute_link_scripts,
    )



from __future__ import annotations
from typing import List, Optional, Set
from rattler.lock.hash import PackageHashes

from rattler.rattler import PyPypiPackageData, PyPypiPackageEnvironmentData


class PypiPackageData:
    """
    A pinned Pypi package.
    """

    _data: PyPypiPackageData

    def satisfies(self, spec: str) -> bool:
        """
        Returns true if this package satisfies the given `spec`.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.satisfies("charset-normalizer")
        True
        >>>
        ```
        """
        return self._data.satisfies(spec)

    @property
    def name(self) -> str:
        """
        The name of the package.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.name
        'charset-normalizer'
        >>>
        ```
        """
        return self._data.name

    @property
    def version(self) -> str:
        """
        The version of the package.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.version
        '3.3.2'
        >>>
        ```
        """
        return self._data.version

    @property
    def url_or_path(self) -> str:
        """
        The URL that points to where the artifact can be downloaded from.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.url_or_path
        'https://files.pythonhosted.org/...'
        >>>
        ```
        """
        return self._data.url_or_path

    @property
    def hash(self) -> Optional[PackageHashes]:
        """
        Hashes of the file pointed to by `url`.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.hash
        PackageHashes()
        >>>
        ```
        """
        return PackageHashes._from_py_package_hashes(self._data.hash)

    @property
    def requires_dist(self) -> List[str]:
        """
        A list of dependencies on other packages.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.requires_dist
        []
        >>>
        ```
        """
        return self._data.requires_dist

    @property
    def requires_python(self) -> Optional[str]:
        """
        The python version that this package requires.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.requires_python
        '>=3.7.0'
        >>>
        ```
        """
        return self._data.requires_python

    @property
    def is_editable(self) -> bool:
        """
        Whether the package should be installed in editable mode or not.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> data = pypi_packages[Platform("osx-arm64")][0][0]
        >>> data.is_editable
        False
        >>>
        ```
        """
        return self._data.is_editable

    @classmethod
    def _from_py_pypi_package_data(cls, pkg_data: PyPypiPackageData) -> PypiPackageData:
        """
        Construct Rattler PypiPackageData from FFI PyPypiPackageData object.
        """
        data = cls.__new__(cls)
        data._data = pkg_data
        return data

    def __repr__(self) -> str:
        """
        Returns a representation of the PypiPackageData.
        """
        return "PypiPackageData()"


class PypiPackageEnvironmentData:
    _data: PyPypiPackageEnvironmentData

    @property
    def extras(self) -> Set[str]:
        """
        The extras enabled for the package.
        Note that the order doesn't matter.

        Examples
        --------
        ```python
        >>> from rattler import LockFile, Platform
        >>> lock_file = LockFile.from_path("../test-data/test.lock")
        >>> env = lock_file.default_environment()
        >>> pypi_packages = env.pypi_packages()
        >>> env_data = pypi_packages[Platform("osx-arm64")][0][1]
        >>> env_data.extras
        set()
        >>>
        ```
        """
        return self._data.extras

    @classmethod
    def _from_py_pypi_env_data(cls, env_data: PyPypiPackageEnvironmentData) -> PypiPackageEnvironmentData:
        """
        Construct Rattler PypiPackageEnvironmentData from FFI PyPypiPackageEnvironmentData object.
        """
        data = cls.__new__(cls)
        data._data = env_data
        return data

    def __repr__(self) -> str:
        """
        Returns a representation of the PypiPackageEnvironmentData.
        """
        return "PypiPackageEnvironmentData()"
